Unit Modules;

{
  Blue Hell software.

  Copyright (C) 1999, 2000 by Jan Punter.
}

Interface

Uses

  SyncObjs, Windows, Messages, MMSystem, SysUtils, Classes, Graphics,
  Controls, Forms, ComCtrls, ExtCtrls, Buttons, Dialogs, Math, Globals;

Const


  UM_PAINT = WM_USER + 0;

Type


  TUMPaint = Packed Record
    Msg      : Cardinal;
    Data     : PRect;
    NotUsed1 : LongInt;
    NotUsed2 : LongInt;
  End;


Const

  DefaultSlotExt = '.Slt'; // Default extension for slot files.
  MaxPoints      = 8192;   // Max nr of points in a graphsheet.

  // Color settings

  Clr1 = clSilver;
  Clr2 = clNavy;

  ModuleColor     : TColor = Clr1;
  ViewerColor     : TColor = Clr2;
  GraphBack       : TColor = clTeal;
  GraphBackPen    : TColor = clGray;
  GraphForePen    : TColor = clYellow;

Var

  GRepeatTimer   : TTimer;
  BmKnob         : TBitMap;
  BmKnobMedium   : TBitMap;
  BmKnobSmall    : TBitMap;
  BmRoundGray    : TBitMap;
  BmRoundYellow  : TBitMap;
  BmRoundRed     : TBitMap;
  BmRoundBlue    : TBitMap;
  BmSquareGray   : TBitMap;
  BmSquareYellow : TBitMap;
  BmSquareRed    : TBitMap;
  BmSquareBlue   : TBitMap;
  BmLedOff       : TBitMap;
  BmLedOn        : TBitMap;

const

  InitRepeatPause = 400;  { repeat times (ms) for knob inc/dec operations }
  RepeatPause     = 100;
  FastRepeatPause =  50;

Type

  TValue     = Integer;
  TShortMidi = DWORD;
  TLongMidi = Record
    Size : Integer;
    Data : Record {Packed Array[ 0 .. Size - 1] Of Byte} End;
  End;
  PLongMidi = ^TLongMidi;


    procedure   Register;
    Function    IsNearNullPoint( aPoint: TPoint): Boolean;
    Function    AssembleShortMidi( aCh, anEvent, aData1, aData2: Byte): TShortMidi;
    Procedure   DisAssembleShortMidi( aShortMidi: TShortMidi; Var aCh, anEvent, aData1, aData2: Byte);
    Function    CreateBitmapFromFile( Const aFileName: String): TBitmap;
    Procedure   Clip( Var aValue: TValue; aLow, aHigh: TValue);


Type

  TConnectorType = (
    ctSymmetric,
    ctAsymmetric,
    ctLogic,
    ctNeuron
  );

  TBezierPoints = Array[ 0 .. 3] Of TPoint;

  EModule       = Class( Exception);
  EBounds       = Class( EModule);
  ECorruptPatch = Class( EModule);
  EInvalidName  = Class( EModule);

  TConnector = Class;

  TNotifyCountChange      = Procedure( aSender: TObject; aCount: Integer)                   Of Object;
  TOnValueValueChange     = Procedure( aSender: TObject; aValue: TValue )                   Of Object;
  TOnBooleanValueChange   = Procedure( aSender: TObject; aValue: Boolean)                   Of Object;
  TNotifyConnectionChange = Procedure( aSender: TObject; aSource, aDestination: TConnector) Of Object;
  TNotifyActiveSlotChange = Procedure( aSender: TObject; Const aSlotName: String)           Of Object;
  TMidiInShortProc        = Procedure( aSender: TObject; aShortMidi: TShortMidi)            Of Object;
  TMidiInLongProc         = Procedure( aSender: TObject; Const aLongMidi: TLongMidi)        Of Object;
  TMidiOutShortProc       = Procedure( aSender: TObject; aShortMidi: TShortMidi)            Of Object;
  TMidiOutLongProc        = Procedure( aSender: TObject; Const aLongMidi: TLongMidi)        Of Object;
  TClickEvent             = Procedure( aSender: TObject; Shift: TShiftState)                Of Object;


  TCustomDetail = Class( TComponent)
  Private
      FLeft         : Integer;
      FTop          : Integer;
      FWidth        : Integer;
      FHeight       : Integer;
      FSnap         : TPoint;
      FPen          : TPen;
      FBrush        : TBrush;
      FHandlesMouse : Boolean;
      FUpdateReq    : Boolean;
    Procedure   DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Virtual;
    Procedure   DoMouseUp  ( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Virtual;
    procedure   DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer);                       Virtual;
    Procedure   SendDblClick  ( Sender: TObject);
    Procedure   SendRightClick( Sender: TObject);
    Function    CenterPoint: TPoint;
    Procedure   InvalidateRectangle( aRect: TRect);                    Virtual;
    Procedure   SetSizeAndPos( R: TRect);                              Virtual;
    Procedure   SetLeft  ( aValue: Integer);
    Procedure   SetTop   ( aValue: Integer);
    Procedure   SetWidth ( aValue: Integer);
    Procedure   SetHeight( aValue: Integer);
    Function    GetClientRect: TRect;
    Function    GetParentRect: TRect;
    Function    GetScreenRect: TRect;
    Function    GetClientOrigin: TPoint;
    Procedure   SetColor( aValue: TColor);
    Function    GetColor: TColor;
    Procedure   SetPen  ( aValue: TPen);
    Procedure   SetBrush( aValue: TBrush);
  Protected
    Procedure   BeginMove ( aPoint: TPoint);                           Virtual;
    Procedure   EndMove   ( aPoint: TPoint);                           Virtual;
    Procedure   UpdateMove( aPoint: TPoint);                           Virtual;
    Procedure   Invalidate;                                            Virtual;
    Procedure   PaintOn( aCanvas: TCanvas);                            Virtual;
    Function    ContainsPoint( aPoint: TPoint): Boolean;               Virtual;
    Function    ClientToScreen( aPoint: TPoint): TPoint;               Virtual;
    Function    ScreenToClient( aPoint: TPoint): TPoint;               Virtual;
    Procedure   SetParentComponent( aValue: TComponent);               Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Function    UniqueName: String;                                    Virtual;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Virtual;
    Function    GetParentComponent: TComponent;                        Override;
    Function    HasParent: Boolean;                                    Override;
    Procedure   VisualMaintainance;                                    Virtual;
  Public
    Property    ClientRect   : TRect         Read GetClientRect;
    Property    ParentRect   : TRect         Read GetParentRect;
    Property    ScreenRect   : TRect         Read GetScreenRect;
    Property    ClientOrigin : TPoint        Read GetClientOrigin;
    Property    Width        : Integer       Read FWidth          Write SetWidth;
    Property    Height       : Integer       Read FHeight         Write SetHeight;
    Property    Brush        : TBrush        Read FBrush          Write SetBrush;
    Property    Pen          : TPen          Read FPen            Write SetPen;
    Property    Color        : TColor        Read GetColor        Write SetColor;
  Published
    Property    Left         : Integer       Read FLeft           Write SetLeft;
    Property    Top          : Integer       Read FTop            Write SetTop;
  End;

  TDetailClass  = Class Of TCustomDetail;

  // a semi-abstract base type for details without mouse interaction
  // (such as TextLabels)
  //
  TNoMouse = Class( TCustomDetail)
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  End;

  // a semi-abstract base type for all connectors.
  TConnector = Class( TCustomDetail)
  Private
      FConnectionCount : Integer;
      FValue           : TValue;
      FConnectorType   : TConnectorType;
  Protected
    Procedure   SetValue        ( aValue: TValue        );
    Procedure   SetConnectorType( aValue: TConnectorType);
    Function    GetConnected: Boolean;
    Procedure   EndMove   ( aPoint: TPoint);                           Override;
    Procedure   UpdateMove( aPoint: TPoint);                           Override;
    Function    GetBitmap: TBitmap;                                    Virtual;
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    Procedure   Connect( aConnector: TConnector);                      Virtual;
    Function    CanAccept( aConnector: TConnector): Boolean;           Virtual;
    Function    SearchDestination: TConnector;
    Procedure   ValueChanged( Var aValue: TValue);                     Virtual;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Procedure   IncConnectionCount;
    Procedure   DecConnectionCount;
  Public
    Property    Value         : TValue         Read FValue         Write SetValue;
    Property    Connected     : Boolean        Read GetConnected;
    Property    ConnectorType : TConnectorType Read FConnectorType Write SetConnectorType;
  End;

  TInput = Class( TConnector)
  Protected
    Function    GetBitmap: TBitmap;                                    Override;
    Function    CanAccept( aConnector: TConnector): Boolean;           Override;
  End;

  TOutput = Class( TConnector)
  Protected
    Function    GetBitmap: TBitmap;                                    Override;
    Function    CanAccept( aConnector: TConnector): Boolean;           Override;
  End;

  TTextLabel = Class( TNoMouse)
  Private
      FText        : String;
      FFont        : TFont;
      FAutoSize    : Boolean;
      FParentColor : Boolean;
    Procedure   AdjustBounds;
    Procedure   SetText( Const aValue: String);
    Procedure   SetTextColor( aValue: TColor);
    Function    GetTextColor: TColor;
    Procedure   SetFont( aValue: TFont);
    Procedure   SetAutoSize( aValue: Boolean);
    Procedure   SetParentColor( aValue: Boolean);
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
  Public
    Property    Text        : String  Read FText        Write SetText;
    Property    TextColor   : TColor  Read GetTextColor Write SetTextColor;
    Property    Font        : TFont   Read FFont        Write SetFont;
    Property    AutoSize    : Boolean Read FAutoSize    Write SetAutoSize;
    Property    ParentColor : Boolean Read FParentColor Write SetParentColor;
  End;

  TViewer = Class( TCustomDetail)
  Private
      FText : String;
      FFont : TFont;
    Procedure   SetText( Const aValue: String);
    Procedure   SetBorderColor( aValue: TColor);
    Function    GetBorderColor: TColor;
    Procedure   SetTextColor( aValue: TColor);
    Function    GetTextColor: TColor;
    Procedure   SetFont( aValue: TFont );
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   VisualMaintainance;                                    Override;
  Public
    Property    Text        : String Read FText          Write SetText;
    Property    BorderColor : TColor Read GetBorderColor Write SetBorderColor;
    Property    TextColor   : TColor Read GetTextColor   Write SetTextColor;
    Property    Font        : TFont  Read FFont          Write SetFont;
  End;

  TKnobMode = (
    kmNone,
    kmTurn,
    kmInc,
    kmDec
  );

  TKnob = Class( TCustomDetail)
  Private
      FValue         : TValue;
      FMinValue      : TValue;
      FMaxValue      : TValue;
      FViewer        : TViewer;
      FOnValueChange : TOnValueValueChange;
      FOldP1         : TPoint;
      FOldP2         : TPoint;
      FKnobMode      : TKnobMode;
      FRepeatCount   : Integer;
    Procedure   DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Override;
    Procedure   DoTimerFired( aSender: TObject);
    Procedure   UpdatePeers;
    Procedure   SetValue   ( aValue: TValue);
    Procedure   SetMinValue( aValue: TValue);
    Procedure   SetMaxValue( aValue: TValue);
    Procedure   SetViewer  ( aValue: TViewer);
    Procedure   SetOnValueChange( aValue: TOnValueValueChange);
  Protected
    Procedure   BeginMove ( aPoint: TPoint);                           Override;
    Procedure   EndMove( aPoint: TPoint);                              Override;
    Procedure   UpdateMove( aPoint: TPoint);                           Override;
    Function    GetBitmap: TBitmap;                                    Virtual;
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Procedure   Invalidate;                                            Override;
  Public
    Property    Value        : TValue              Read FValue         Write SetValue;
    Property    MinValue     : TValue              Read FMinValue      Write SetMinValue;
    Property    MaxValue     : TValue              Read FMaxValue      Write SetMaxValue;
    Property    Viewer       : TViewer             Read FViewer        Write SetViewer;
    Property    KnobMode     : TKnobMode           Read FKnobMode;
    Property    OnValueChange: TOnValueValueChange Read FOnValueChange Write SetOnValueChange;
  End;

  TKnobMedium = Class( TKnob)
  Protected
    Function    GetBitmap: TBitmap;                                    Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  End;

  TKnobSmall = Class( TKnob)
  Protected
    Function    GetBitmap: TBitmap;                                    Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  End;

  TExpKnob = Class( TKnob)
  Protected
    Procedure   EndMove( aPoint: TPoint);                              Override;
    Procedure   UpdateMove( aPoint: TPoint);                           Override;
  End;

  TPushButton = Class( TCustomDetail)
  Private
      FDown          : Boolean;
      FOldDown       : Boolean;
      FBitmap        : TBitmap;
      FOnValueChange : TOnBooleanValueChange;
      FGroupId       : Integer;
      FAllowAllUp    : Boolean;
      FOnClick       : TNotifyEvent;
    Procedure   SelectUnique;
    Procedure   SetBitmap( aValue: TBitmap);
    Procedure   SetDown( aValue: Boolean);
    Procedure   SetGroupId( aValue: Integer);
    Procedure   SetAllowAllUp( aValue: Boolean);
    Procedure   SetOnValueChange( aValue: TOnBooleanValueChange);
  Protected
    Procedure   BeginMove ( aPoint: TPoint);                           Override;
    Procedure   EndMove   ( aPoint: TPoint);                           Override;
    Procedure   UpdateMove( aPoint: TPoint);                           Override;
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   LoadBitmap( Const aFileName: String);
  Protected
    Property    Bitmap: TBitmap Read FBitmap Write SetBitmap;
  Public
    Property    Down         : Boolean               Read FDown          Write SetDown;
    Property    GroupId      : Integer               Read FGroupId       Write SetGroupId;
    Property    AllowAllUp   : Boolean               Read FAllowAllUp    Write SetAllowAllUp;
    Property    OnValueChange: TOnBooleanValueChange Read FOnValueChange Write SetOnValueChange;
    Property    OnClick      : TNotifyEvent          Read FOnClick       Write FOnClick;
  End;

  TIllustration = Class( TNoMouse)
  Private
      FBitmap      : TBitmap;
      FTransparent : Boolean;
      FAutoSize    : Boolean;
    Procedure   AdjustBounds;
    Procedure   SetBitmap     ( aValue: TBitmap);
    Procedure   SetTransparent( aValue: Boolean);
    Procedure   SetAutoSize   ( aValue: Boolean);
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   LoadBitmap( Const aFileName: String);
  Public
    Property    Bitmap      : TBitmap Read FBitmap      Write SetBitmap;
    Property    Transparent : Boolean Read FTransparent Write SetTransparent;
    Property    AutoSize    : Boolean Read FAutoSize    Write SetAutoSize;
  End;

  TPointArray = Array[ 0 .. MaxPoints - 1] Of TPoint;
  PPointArray = ^TPointArray;

  TGraphSheet = Class( TNoMouse)
  Private
      FData      : PPointArray;
      FEltCount  : Integer;
      FXScale    : Integer;
      FYScale    : Integer;
      FDrawColor : TColor;
    Procedure   CheckBounds( anIndex: Integer);
    Function    GetData     ( anIndex: Integer): TPoint;
    Procedure   SetData     ( anIndex: Integer; aValue: TPoint);
    Function    GetXData    ( anIndex: Integer): Integer;
    Procedure   SetXData    ( anIndex: Integer; aValue: Integer);
    Function    GetYData    ( anIndex: Integer): Integer;
    Procedure   SetYData    ( anIndex: Integer; aValue: Integer);
    Function    GetXScaled  ( anIndex: Integer): Integer;
    Function    GetYScaled  ( anIndex: Integer): Integer;
    Procedure   SetEltCount ( aValue: Integer);
    Procedure   SetXScale   ( aValue: Integer);
    Procedure   SetYScale   ( aValue: Integer);
    Procedure   SetDrawColor( aValue: TColor );
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Override;
    Procedure   Clear;
    Procedure   VisualMaintainance;                                    Override;
  Public
    Property    Data   [ anIndex: Integer]: TPoint  Read GetData    Write SetData;
    Property    XData  [ anIndex: Integer]: Integer Read GetXData   Write SetXData;
    Property    YData  [ anIndex: Integer]: Integer Read GetYData   Write SetYData;
    Property    XScaled[ anIndex: Integer]: Integer Read GetXScaled;
    Property    YScaled[ anIndex: Integer]: Integer Read GetYScaled;
    Property    EltCount                  : Integer Read FEltCount  Write SetEltCount;
    Property    XScale                    : Integer Read FXScale    Write SetXScale;
    Property    YScale                    : Integer Read FYScale    Write SetYScale;
    Property    DrawColor                 : TColor  Read FDrawColor Write SetDrawColor;
  End;

  TIndicator = Class( TNoMouse)
  Private
      FActive   : Boolean;
    Procedure   SetActive  ( aValue: Boolean);
  Protected
    Function    GetBitmap: TBitmap;                                    Virtual;
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Procedure   VisualMaintainance;                                    Override;
  Public
    Property    Active   : Boolean Read FActive   Write SetActive;
  end;

  TCustomModule = Class( TCustomDetail)
  Private
      FSelected   : Boolean;
      FInCreation : Boolean;
      FTextLabel  : TTextLabel;
    Procedure   DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Override;
    Procedure   DoMouseUp  ( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Override;
    Procedure   DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer);                       Override;
    Procedure   SetSelected( aValue: Boolean);
    Procedure   HandleSelection( CtrlKeyActive: Boolean);
    Function    GetTitle: String;
    Procedure   SetTitle( Const aValue: String);
    Procedure   HandleUpdateMove( aPoint: TPoint);
    Procedure   UnselectPeers;
    Procedure   BringToFront;
    Procedure   Disconnect( aConnector: TConnector);                   Virtual;
    Procedure   MarkConnections;
  Protected
    Procedure   Notification( aComponent: TComponent; anOperation: TOperation); Override;
    Procedure   HandleSelectionChange( aValue: Boolean);               Virtual;
    Procedure   Restack;
    Procedure   EndMove( aPoint: TPoint);                              Override;
    Procedure   UpdateMove( aPoint: TPoint);                           Override;
    Procedure   Move( aPoint: TPoint);                                 Virtual;
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    Procedure   PaintDetails( aCanvas: TCanvas);                       Virtual;
    Procedure   CreateDetails;                                         Virtual;
    Function    CreateDetail( aLeft, aTop: Integer; Const aName: String; aDetailClass: TDetailClass): TCustomDetail;
    Function    SearchDestination( aSource: TConnector; aPoint: TPoint): TConnector;
    Function    SearchDetail( aPoint: TPoint): TCustomDetail;
    Function    FindDestination( aSource: TConnector; aPoint: TPoint): TConnector;
    Procedure   Connect( aSource, aDestination: TConnector);           Virtual;
    Function    ClientToScreen( aPoint: TPoint): TPoint;               Override;
    Function    ScreenToClient( aPoint: TPoint): TPoint;               Override;
    Procedure   ScrollSelfInView;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Execute;                                               Virtual;
    Procedure   VisualMaintainance;                                    Override;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Override;
  Public
    Property    Selected  : Boolean    Read FSelected  Write SetSelected;
    Property    TextLabel : TTextLabel Read FTextLabel Write FTextLabel;
  Published
    Property    Title     : String     Read GetTitle   Write SetTitle;
  End;

  TModuleClass = Class Of TCustomModule;

  TModule = Class( TCustomModule)
  End;

  TModule2 = Class( TModule)
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  End;

  TModule4 = Class( TModule)
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  End;

  TModule8 = Class( TModule)
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  End;

  TMidiModule = Class( TCustomModule)
  Private
  Protected
    Procedure   AcceptMidiShort( aShortMidi: TShortMidi);              Virtual;
    Procedure   AcceptMidiLong( aLongMidi: TLongMidi);                 Virtual;
  Public
    Procedure   MidiOutShort( aShortMidi: TShortMidi);                 Virtual;
    Procedure   MidiOutLong( Const aLongMidi: TLongMidi);              Virtual;
  End;

  TConnection = Class
  Private
      FSource      : TConnector;
      FDestination : TConnector;
      FMarked      : Boolean;
      FSelected    : Boolean;
    Procedure   SetSelected( aValue: Boolean);
  Public
    Constructor Create( aSource, aDestination: TConnector);
    Destructor  Destroy;                                               Override;
  Public
    Property    Source      : TConnector Read FSource;
    Property    Destination : TConnector Read FDestination;
    property    Selected    : Boolean    Read FSelected  Write SetSelected;
  End;

  TDrawSheet = Class;

  TConnections = Class( TList)
  Private
      FOwner : TDrawSheet;
    Function    GetConnectionCount: Integer;
    Function    GetConnection ( anIndex: Integer): TConnection;
    Procedure   Disconnect( aConnector: TConnector);
  Public
    Constructor Create( anOwner: TDrawSheet);
    Function    Connect( aSource, aDestination: TConnector): Boolean;
    Function    FindConnection( aSource, aDestination: TConnector): Integer;
    Procedure   Execute;                                               Virtual;
  Public
    Property    ConnectionCount : Integer Read GetConnectionCount;
    Property    Connection [ anIndex: Integer]: TConnection Read GetConnection;
    Property    Owner: TDrawSheet Read FOwner;
  End;

  TSlot = Class;

  TDrawSheet = Class( TControl)
  Private
      FModules           : TList;
      FConnections       : TConnections;
      FSlot              : TSlot;
      FCanvas            : TCanvas;
      FPen               : TPen;
      FSlotActive        : Boolean;
      FDoDrawConnections : Boolean;
      FConnectionFixups  : TStringList;
    procedure   WMPaint       ( Var aMessage: TWMPaint); Message WM_PAINT;
    Procedure   DoMouseDown   ( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Procedure   DoMouseUp     ( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure   DoMouseMove   ( Sender: TObject; Shift: TShiftState; X, Y: Integer);
    Procedure   SendDblClick  ( Sender: TObject);
    Procedure   SendRightClick( Sender: TObject);

    Procedure   SetPen        ( aValue: TPen   );
    Procedure   SetSlotActive ( aValue: Boolean);
    Function    GetModule     ( anIndex: Integer): TCustomModule;
    Function    GetConnection ( anIndex: Integer): TConnection;
    Function    GetSource     ( anIndex: Integer): TConnector;
    Function    GetDestination( anIndex: Integer): TConnector;
    Procedure   PerformCountChange;
    Procedure   HandleModuleUpdateMove( aModule: TCustomModule; aPoint: TPoint);
    Procedure   HandleModuleMoved( aPoint: TPoint);
    Procedure   UnselectAllModulesBut( aModule: TCustomModule);
    Procedure   Disconnect( aConnector: TConnector);
    Procedure   InvalidateMarkedConnections;
    Procedure   MarkConnections( aConnector: TConnector);
    Procedure   ClearConnectionFixups;
    Procedure   AddConnectionFixup( Const S: String);
    Procedure   FixupConnections;
    Procedure   CreateConnectionFromFixup( Const S: String);
    Procedure   SyncModules;
    Procedure   PerformMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Procedure   PerformMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
  Protected
    Procedure   BeginMove ( aPoint: TPoint);                           Virtual;
    Procedure   EndMove   ( aPoint: TPoint);                           Virtual;
    Procedure   UpdateMove( aPoint: TPoint);                           Virtual;
    Procedure   Notification( aComponent: TComponent; anOperation: TOperation); Override;
    Procedure   BringModuleToFront( aModule: TCustomModule);
    Procedure   SelectModulesInRect( aRect: TRect);
    Procedure   InvalidateRectangleDetail( aRect: TRect);
    Function    SearchDetail( aPoint: TPoint): TCustomDetail;
    Function    ContainsPoint( aPoint: TPoint): Boolean;               Virtual;
    Function    FindDestination( aSource: TConnector; aPoint: TPoint): TConnector;
    Procedure   Connect( aSource, aDestination: TConnector);           Virtual;
    Procedure   Paint;                                                 Virtual;
    Procedure   PaintDetails    ( aCanvas: TCanvas);
    Procedure   PaintConnection ( aCanvas: TCanvas; aConnection: TConnection);
    Procedure   PaintConnections( aCanvas: TCanvas);
    Procedure   ScrollModuleInView( aModule: TCustomModule);
    Procedure   ConnectionMade( aSource, aDestination: TConnector);    Virtual;
    Procedure   ConnectionLost( aSource, aDestination: TConnector);    Virtual;
    Procedure   GetChildren( aProc: TGetChildProc; Root: TComponent);  Override;
    Procedure   DefineProperties( aFiler : TFiler );                   Override;
    Procedure   WriteConnections( aWriter: TWriter);
    Procedure   ReadConnections ( aReader: TReader);
    Procedure   Loaded;                                                Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Execute;                                               Virtual;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Virtual;
    Procedure   AcceptMidiShort( aShortMidi: TShortMidi);
    Procedure   AcceptMidiLong( Const aLongMidi: TLongMidi);
    Function    CreateModule( aModuleClass: TModuleClass): TCustomModule;
    Procedure   FreeModule( anIndex: Integer);
    Procedure   FreeModules;
    Procedure   FreeSelectedModules;
    Procedure   CopyModulesToClipBoard;
    Procedure   PasteModulesFromClipBoard;
    Procedure   SelectAllModules;
    Procedure   InvertModuleSelection;
    Function    IndexOfModule( aModule: TCustomModule): Integer;
    Function    ModuleCount    : Integer;
    Function    ConnectionCount: Integer;
    Procedure   VisualMaintainance;                                    Virtual;
  Public
    Property    Pen        : TPen    Read FPen        Write SetPen;
    Property    SlotActive : Boolean Read FSlotActive Write SetSlotActive;
    Property    Module     [ anIndex: Integer]: TCustomModule  Read GetModule;
    Property    Connection [ anIndex: Integer]: TConnection    Read GetConnection;
    Property    Source     [ anIndex: Integer]: TConnector     Read GetSource;
    Property    Destination[ anIndex: Integer]: TConnector     Read GetDestination;
  End;

  TSlot = Class( TScrollBox)
  Private
      FOnRightClick       : TNotifyEvent;
      FOnCountChange      : TNotifyCountChange;
      FPrivateCountChange : TNotifyCountChange;
      FOnConnectionMade   : TNotifyConnectionChange;
      FOnConnectionLost   : TNotifyConnectionChange;
      FOnMidiOutShort     : TMidiOutShortProc;
      FOnMidiOutLong      : TMidiOutLongProc;
      FDrawSheet          : TDrawSheet;
      FFileName           : String;
      FActive             : Boolean;
    Procedure   PerformCountChange;
    Procedure   SendDblClick  ( Sender: TObject);
    Procedure   SendRightClick( Sender: TObject);
    Procedure   UMPaint( Var aMessage: TUMPaint);              Message UM_PAINT;
    Procedure   PerformMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Procedure   PerformMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
    Procedure   SetActive( aValue: Boolean);
  Protected
    Procedure   ConnectionMade( aSource, aDestination: TConnector);    Virtual;
    Procedure   ConnectionLost( aSource, aDestination: TConnector);    Virtual;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Execute;                                               Virtual;
    Procedure   Save;                                                  Virtual;
    Procedure   SaveAs( Const aFileName: String);                      Virtual;
    Procedure   Load( Const aFileName: String);                        Virtual;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Virtual;
    Procedure   AcceptMidiShort( aShortMidi: TShortMidi);
    Procedure   AcceptMidiLong( Const aLongMidi: TLongMidi);
    Function    CreateModule( aModuleClass: TModuleClass): TCustomModule;
    Procedure   FreeModule( anIndex: Integer);
    Procedure   FreeModules;
    Procedure   FreeSelectedModules;
    Procedure   CopyModulesToClipBoard;
    Procedure   PasteModulesFromClipBoard;
    Procedure   SelectAllModules;
    Procedure   InvertModuleSelection;
    Function    ModuleCount: Integer;
    Procedure   ScrollModuleInView( aModule: TCustomModule);
    Procedure   VisualMaintainance;                                    Virtual;
  Protected
    Property    OnPrivateCountChange : TNotifyCountChange Read FPrivateCountChange Write FPrivateCountChange;
  Published
    Property    OnRightClick     : TNotifyEvent            Read FOnRightClick     Write FOnRightClick;
    Property    OnCountChange    : TNotifyCountChange      Read FOnCountChange    Write FOnCountChange;
    Property    OnConnectionMade : TNotifyConnectionChange Read FOnConnectionMade Write FOnConnectionMade;
    Property    OnConnectionLost : TNotifyConnectionChange Read FOnConnectionLost Write FOnConnectionLost;
    Property    OnMidiOutShort   : TMidiOutShortProc       Read FOnMidiOutShort   Write FOnMidiOutShort;
    Property    OnMidiOutLong    : TMidiOutLongProc        Read FOnMidiOutLong    Write FOnMidiOutLong;
    Property    SlotFileName     : String                  Read FFileName;
    Property    Active           : Boolean                 Read FActive           Write SetActive;
  End;

  TModulePageControl = Class( TPageControl)
  Private
      FOnSlotCountChange   : TNotifyCountChange;
      FOnModuleCountChange : TNotifyCountChange;
      FOnConnectionMade    : TNotifyConnectionChange;
      FOnConnectionLost    : TNotifyConnectionChange;
      FOnActiveSlotChange  : TNotifyActiveSlotChange;
      FOnMidiOutShort      : TMidiOutShortProc;
      FOnMidiOutLong       : TMidiOutLongProc;
      FOnDetailClick       : TNotifyEvent;
      FOnDetailRightClick  : TNotifyEvent;
      FOnDetailDblClick    : TNotifyEvent;
      FPageCount           : Integer;
      FLockCount           : Integer;
    Function    GetActiveSlot: TSlot;
    Procedure   DoPrivateCountChange( Sender: TObject; aCount: Integer);
    Procedure   DoPrivateConnectionMade( aSender: TObject; aSource, aDestination: TConnector);
    Procedure   DoPrivateConnectionLost( aSender: TObject; aSource, aDestination: TConnector);
    Procedure   DoMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Procedure   DoMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
    Procedure   PerformSlotCountChange;
    Procedure   PerformModuleCountChange( Sender: TSlot);
    Procedure   PerformConnectionMade( aSender: TObject; aSource, aDestination: TConnector);
    Procedure   PerformConnectionLost( aSender: TObject; aSource, aDestination: TConnector);
    Procedure   PerformSlotChange( aSender: TObject);
    Procedure   PerformClick     ( aSender: TObject);
    Procedure   PerformRightClick( aSender: TObject);
    Procedure   PerformDblClick  ( aSender: TObject);
    Procedure   Lock;
    Procedure   UnLock;
    Function    GetCurvedLines: Boolean;
    Procedure   SetCurvedLines( aValue: Boolean);
  Protected
    Procedure   Change;                                                Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Procedure   Execute;                                               Virtual;
    Procedure   Save;                                                  Virtual;
    Procedure   SaveAll;                                               Virtual;
    Procedure   SaveAs( Const aFileName: String);                      Virtual;
    Procedure   Load( Const aFileName: String);                        Virtual;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Virtual;
    Procedure   AddSlot( Const aName: String);
    Procedure   AcceptMidiShort( aShortMidi: TShortMidi);
    Procedure   AcceptMidiLong( Const aLongMidi: TLongMidi);
    Procedure   FreeCurrentSlot;
    Procedure   FreeAllSlots;
    Function    CreateModule( aModuleClass: TModuleClass): TCustomModule;
    Procedure   FreeModule( anIndex: Integer);
    Procedure   FreeModules;
    Procedure   FreeSelectedModules;
    Procedure   CopyModulesToClipBoard;
    Procedure   PasteModulesFromClipBoard;
    Procedure   SelectAllModules;
    Procedure   InvertModuleSelection;
    Procedure   ShakeCables;
    Function    ModuleCount : Integer;
    Procedure   VisualMaintainance;                                    Virtual;
  Public
    Property    ActiveSlot: TSlot Read GetActiveSlot;
  Published
    Property    CurvedLines         : Boolean                 Read GetCurvedLines       Write SetCurvedLines;
    Property    OnSlotCountChange   : TNotifyCountChange      Read FOnSlotCountChange   Write FOnSlotCountChange;
    Property    OnModuleCountChange : TNotifyCountChange      Read FOnModuleCountChange Write FOnModuleCountChange;
    Property    OnConnectionMade    : TNotifyConnectionChange Read FOnConnectionMade    Write FOnConnectionMade;
    Property    OnConnectionLost    : TNotifyConnectionChange Read FOnConnectionLost    Write FOnConnectionLost;
    Property    OnActiveSlotChange  : TNotifyActiveSlotChange Read FOnActiveSlotChange  Write FOnActiveSlotChange;
    Property    OnMidiOutShort      : TMidiOutShortProc       Read FOnMidiOutShort      Write FOnMidiOutShort;
    Property    OnMidiOutLong       : TMidiOutLongProc        Read FOnMidiOutLong       Write FOnMidiOutLong;
    Property    OnDetailClick       : TNotifyEvent            Read FOnDetailClick       Write FOnDetailClick;
    Property    OnDetailRightClick  : TNotifyEvent            Read FOnDetailRightClick  Write FOnDetailRightClick;
    Property    OnDetailDblClick    : TNotifyEvent            Read FOnDetailDblClick    Write FOnDetailDblClick;
  End;

  TModuleButton = Class( TSpeedButton)
  Private
      FPageName    : String;
      FModuleClass : TModuleClass;
  Public
    Procedure   LoadGlyph;
    Procedure   Initialize( Const aPageName: String; aModuleClass: TModuleClass);
  Public
    Property    PageName   : String       Read FPageName;
    Property    ModuleClass: TModuleClass Read FModuleClass;
  End;

  TModuleButtonClick = Procedure( aSender: TObject; aModuleClass: TModuleClass) Of Object;

  TmoduleSelector = Class( TPageControl)
  Private
      FRegisteredModules   : TList;
      FOnModuleButtonClick : TModuleButtonClick;
      FButtonsEnabled      : Boolean;
    Procedure   SetButtonsEnabled( aValue: Boolean);
    Function    AddPage( Const aName: String): TTabSheet;
    Function    PageButtonCount( Const aName: String): Integer;
    Function    PageExists( Const aName: String): Boolean;
    Function    PageByName( Const aName: String): TTabSheet;
    Function    FindModulePage( aModuleClass: TModuleClass): TTabSheet;
    Function    FindModuleButton( aModuleClass: TModuleClass): TModuleButton;
    Procedure   DoModuleMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
    Procedure   FormatPage( aPage: TTabSheet);
  Protected
    Procedure   Notification( aComponent: TComponent; anOperation: TOperation); Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   RegisterModuleClass( Const aPageName: String; aModuleClass: TModuleClass);
    Procedure   UnregisterModuleClass( aModuleClass: TModuleClass);
  Published
    Property    OnModuleButtonClick : TModuleButtonClick Read FonModuleButtonClick Write FOnModuleButtonClick;
    Property    ButtonsEnabled      : Boolean            Read FButtonsEnabled      Write SetButtonsEnabled;
  End;




implementation

Uses

  ClipBrd;


Const

  NullPoint : TPoint = ( X:0; Y:0);

  procedure   Register;
  begin
    RegisterComponents(
      'Modules',
      [
        TModuleSelector,
        TModulePageControl
      ]
    );
  end;

  Function IsNearNullPoint( aPoint: TPoint): Boolean;
  Begin
    With aPoint Do
      Result := ( Abs( x) < 10) And ( Abs( y) < 10);
  End;

  Function  AssembleShortMidi( aCh, anEvent, aData1, aData2: Byte): TShortMidi;
  Begin
    Result := Integer( aCh And $0f)+ Integer( anEvent And $0f) Shl 4 + Integer( aData1) Shl 8 + Integer( aData2) Shl 16;
  End;

  Procedure DisAssembleShortMidi( aShortMidi: TShortMidi; Var aCh, anEvent, aData1, aData2: Byte);
  Begin
    aCh     := ( aShortMidi       ) And $0f;
    anEvent := ( aShortMidi Shr  4) And $0f;
    aData1  := ( aShortMidi Shr  8) And $7f;
    aData2  := ( aShortMidi Shr 16) And $7f;
  End;

  Function CreateBitmapFromFile( Const aFileName: String): TBitmap;
  Begin
    Result := TBitmap.Create;
    With Result Do
    Begin
      LoadFromFile( ApplicationPath + 'bitmaps\' + aFileName);
      Transparent     := True;
      TransparentMode := tmAuto;
    End;
  End;

  Procedure   Clip( Var aValue: TValue; aLow, aHigh: TValue);
  Begin
    If aValue < aLow
    Then aValue := aLow
    Else If aValue > aHigh
    Then aValue := aHigh;
  End;


// Utility functions.

    Function  CleanFileName( Const aFileName: String): String;
    Var
      p : Integer;
    Begin
      Result := aFileName;
      p := Pos( Uppercase( DefaultSlotExt), UpperCase( Result));
      If p > 0
      Then Result := Copy( Result, 1, p - 1);
      p := Length( Result);
      While p > 0 Do
        If Result[ p] = '\'
        Then Break
        Else Dec( p);
      If p < Length( Result)
      Then Result := Copy( Result, p + 1, Length( Result));
    End;

    Procedure Snap( Var aPoint: TPoint; aSnap: TPoint);
    Begin
      With aSnap Do
      Begin
        If x <>  0
        Then aPoint.x := (( aPoint.x + x Div 2) Div x) * x;
        If y <> 0
        Then aPoint.y := (( aPoint.y + y Div 2) Div y) * y;
      End;
    End;

    Function  Indent( anInd: Integer; S: String): String;
    Const
      IStr        : String = ' ';
      Initialized : Boolean = False;
    Begin
      If Not Initialized
      Then Begin
        While Length( IStr) < 128 Do
          IStr := IStr + IStr;
        Initialized := True;
      End;
      Result := Format( ' %s%s', [ Copy( IStr, 1, 2 * anInd), S]);
    End;

    Function   RectanglesOverlap( A, B: TRect): Boolean;
    Var
      R : TRect;
    Begin
      Result := IntersectRect( R, A, B);
    End;

    Procedure  Exchange( Var A, B: Integer);
    Var
      T : Integer;
    Begin
      T := A;
      A := B;
      B := T;
    End;

    Function   AddPoints( A, B: TPoint): TPoint;
    Begin
      With Result Do
      Begin
        x := A.x + B.x;
        y := A.y + B.y;
      End;
    End;

    Function   SubPoints( A, B: TPoint): TPoint;
    Begin
      With Result Do
      Begin
        x := A.x - B.x;
        y := A.y - B.y;
      End;
    End;

    Function   MultPoint( A: TPoint; M: Integer): TPoint;
    Begin
      Result := Point( A.x * M, A.y * M);
    End;

    Function   DivPoint( A: TPoint; D: Integer): TPoint;
    Begin
      Result := Point( A.x Div D, A.y Div D);
    End;

    Function MoveRect( aRect: TRect; aPoint: TPoint): TRect;
    Begin
      Result := aRect;
      With Result, aPoint Do
      Begin
        Left   := Left   + x;
        Top    := Top    + y;
        Right  := Right  + x;
        Bottom := Bottom + y;
      End;
    End;

    Function InflateRectangle( R: TRect; a: Integer): TRect;
    Begin
      Result := R;
      With Result Do
      Begin
        Dec( Left  , a);
        Dec( Top   , a);
        Inc( Right , a);
        Inc( Bottom, a);
      End;
    End;

    // For angles we we use the mathematical vision, 0 degrees point to the right.

    Function  ValuesToAngle( MinValue, Value, MaxValue: TValue): Integer;
    Begin
      If MinValue <> MaxValue
      Then Result := 270 * ( Value - MinValue) Div ( MinValue - MaxValue) + 225
      Else Result := 0; // arbitrary ...
    End;

    Function  PolarToPoint( aCenter: TPoint; anAngle, aLength: Integer): TPoint;
    Begin
      With Result Do
      Begin
        x := Round( aCenter.x + aLength * Cos( DegToRad( anAngle)));
        y := Round( aCenter.y - aLength * Sin( DegToRad( anAngle)));
      End;
    End;

    Function  DistanceToValue( MinValue, MaxValue: TValue; aPoint: TPoint): TValue;
    Var
      anAngle: Extended;
    Begin
      anAngle := RadToDeg( ArcTan2( - aPoint.y, aPoint.x));
      While anAngle < 0 Do
        anAngle := anAngle + 360;
      While anAngle > 270 Do
        anAngle := anAngle - 360;
      Result := MinValue + Round( (MinValue - MaxValue) * ( anAngle - 225) / 270);
      If MaxValue > MinValue
      Then Begin
        If Result > MaxValue
        Then Result := MaxValue
        Else If Result < MinValue
        Then Result := MinValue;
      End
      Else begin
        If Result > MinValue
        Then Result := MinValue
        Else If Result < MaxValue
        Then Result := MaxValue;
      End
    End;

    Function  DistanceToExpValue( MinValue, MaxValue: TValue; aPoint: TPoint): TValue;
    Var
      anAngle: Extended;
    Begin
      anAngle := RadToDeg( ArcTan2( - aPoint.y, aPoint.x));
      While anAngle < 0 Do
        anAngle := anAngle + 360;
      While anAngle > 270 Do
        anAngle := anAngle - 360;
      Result := MinValue + Round( (MinValue - MaxValue) * ( Power( 1000, ( anAngle - 225) / 270) - 1) / 999);
      If MaxValue > MinValue
      Then Begin
        If Result > MaxValue
        Then Result := MaxValue
        Else If Result < MinValue
        Then Result := MinValue;
      End
      Else begin
        If Result > MinValue
        Then Result := MinValue
        Else If Result < MaxValue
        Then Result := MaxValue;
      End
    End;

    Function   Rotate( A: TPoint; angle: Integer): TPoint;
    Var
      Ang : Extended;
    Begin
      Ang := DegToRad( angle);
      With Result Do
      Begin
        x := Round( Cos( Ang) * A.x - Sin( Ang) * A.y);
        y := Round( Sin( Ang) * A.x + Cos( Ang) * A.y);
      End;
    End;


Var

  GUseCurvedLines : Boolean = False;
  GBezierAngle1   : Integer = 20;
  GBezierStrength1: Integer = 30;
  GBezierAngle2   : Integer = 30;
  GBezierStrength2: Integer = 50;

  Procedure   CalcBezierCurve( Var P: TBezierPoints);
  Var
    Dir1 : TPoint;
    Dir2 : TPoint;
  Begin
    If GUseCurvedLines
    Then Begin
      Dir1  := Rotate( SubPoints( P[ 0], P[ 3]), GBezierAngle1);
      Dir2  := Rotate( SubPoints( P[ 0], P[ 3]), GBezierAngle2);
      P[ 1] := SubPoints( p[ 0], DivPoint( MultPoint( Dir1, GBezierStrength1), 100));
      P[ 2] := AddPoints( p[ 3], DivPoint( MultPoint( Dir2, GBezierStrength2), 100));
    End
    Else Begin
      P[ 1] := p[ 3];
      P[ 2] := P[ 0];
    End;
  End;

    Function  LinesIntersect( A, B, C, D: TPoint): Boolean;
    //
    // Equations :
    //
    //   l1 :: A + p * ( B - A)
    //   l2 :: C + q + ( D - C)
    //
    //   The lines l1 and l2 intersect when the equation l1 = l2 has
    //   a solution and the point of intersection is between the end
    //   points of both of the lines.
    //
    //   l1 = l2 ->
    //   A + p * ( B - A) = C + q * ( D - C)
    //
    //   p(Bx-Ax) + q(Cx-Dx) = Cx-Ax
    //   p(By-Ay) + q(Cy-Dy) = Cy-Ay
    //
    //   Delta  = (Bx-Ax)(Cy-Dy) - (By-Ay)(Cx-Dx)
    //   Delta1 = (Cx-Ax)(Cy-Dy) - (Cy-Ay)(Cx-Dx)
    //   Delta2 = (Bx-Ax)(Cy-Dy) - (By-Ay)(Cx-Ax)
    //
    //   We have no sulutions if Delta = 0 And ( Delta1 <> 0 Or Delta2 <> 0),
    //   otherwise we have one solution or ininitly many; when we have one solution
    //   we will have to check if this solution falls between the end points
    //   of any of the lines. The solution is the point
    //   (Delta1/Delta, Delta2/Delta)
    Var
      DxBA   : Integer;
      DyBA   : Integer;
      DxCD   : Integer;
      DyCD   : Integer;
      DxCA   : Integer;
      DyCA   : Integer;
      Delta  : Integer;
      Delta1 : Integer;
      Delta2 : Integer;
      p      : Extended;
      q      : Extended;
      Sx1    : Extended;
      Sx2    : Extended;
    Begin
      DxBA   := B.x - A.x;
      DyBA   := B.y - A.y;
      DxCD   := C.x - D.x;
      DyCD   := C.y - D.y;
      DxCA   := C.x - A.x;
      DyCA   := C.y - A.y;
      Delta  := DxBA * DyCD - DyBA * DxCD;
      Delta1 := DxCA * DyCD - DyCA * DxCD;
      Delta2 := DxBA * DyCD - DyBA * DxCA;
      If Delta = 0
      Then Result := ( Delta1 = 0) And ( Delta2 = 0) // this is not ok, we need to check overlap ...
      Else Begin
        p := Delta1 / Delta;
        q := Delta2 / Delta;
        Sx1 := A.x + p * (B.x - A.x);
        Sx2 := C.x + q * (D.x - C.x);
        If A.x > B.x
        Then Exchange( A.x, B.x);
        If C.x > D.x
        Then Exchange( C.x, D.x);
        Result := ( Sx1 >= A.x) And ( Sx1 <= B.x) And (Sx2 >= C.x) And (Sx2 <= D.x);
      End;
    End;

    Function  LineIntersectsRectangle( A, B: TPoint; R: TRect): Boolean;
    //
    //  A line intersects a rectangle when it intersects the rectangles
    //  bounding box, or when the line is somehow contained within
    //  the the rectangle (that is A or B are in R, well actualy for
    //  containment A and B both have to be within R, but A or B is
    //  faster check (and the concern here is not proper contaimnent
    //  per se)).
    //
    Begin
      With R Do
        Result :=
          // Containment of A or B in R
          (
            PtInRect( R, A) Or
            PtInRect( R, B)
          )
          Or
          // Intersection of the bounding box
          (
            LinesIntersect( A, B, Point( Left , Top   ), Point( Right, Top   )) Or
            LinesIntersect( A, B, Point( Right, Top   ), Point( Right, Bottom)) Or
            LinesIntersect( A, B, Point( Right, Bottom), Point( Left , Bottom)) Or
            LinesIntersect( A, B, Point( Left , Bottom), Point( Left , Top   ))
          );
    End;


    Function  BezierIntersectsRectangle( aBezierPoints: TBezierPoints; R: TRect): Boolean;
    //
    //  A Bezier intersects a rectangle when ?
    Begin
      Result := True;
    End;




Var

  GAnchor       : TPoint     = ( x: 0; y: 0);
  GDistance     : TPoint     = ( x: 0; y: 0);
  GDragged      : TComponent = Nil;
  GLastDragged  : TComponent = Nil;
  GAutoCreation : Boolean    = False;



{ ========
  TCustomDetail = Class( TControl)
  Private
      FLeft         : Integer;
      FTop          : Integer;
      FWidth        : Integer;
      FHeight       : Integer;
      FSnap         : TPoint;
      FPen          : TPen;
      FBrush        : TBrush;
      FHandlesMouse : Boolean;
      FUpdateReq    : Boolean;
  Public
    Property    ClientRect   : TRect         Read GetClientRect;
    Property    ParentRect   : TRect         Read GetParentRect;
    Property    ScreenRect   : TRect         Read GetScreenRect;
    Property    ClientOrigin : TPoint        Read GetClientOrigin;
    Property    Width        : Integer       Read FWidth          Write SetWidth;
    Property    Height       : Integer       Read FHeight         Write SetHeight;
    Property    Brush        : TBrush        Read FBrush          Write SetBrush;
    Property    Pen          : TPen          Read FPen            Write SetPen;
    Property    Color        : TColor        Read GetColor        Write SetColor;
  Published
    Property    Left         : Integer       Read FLeft           Write SetLeft;
    Property    Top          : Integer       Read FTop            Write SetTop;
  Private
}

    Procedure   TCustomDetail.DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Virtual;
    Begin
      If ssDouble In Shift
      Then SendDblClick( Self)
      Else If Button = mbRight
      Then SendRightClick( Self)
      Else If ( Button = mbLeft) And ( GDragged = Nil)
      Then Begin
        GDragged     := Self;
        GLastDragged := GDragged;
        GAnchor      := CenterPoint;
        GDistance    := Point( 0, 0);
        BeginMove ( GAnchor  );
        UpdateMove( GDistance);
      End;
    End;

    Procedure   TCustomDetail.DoMouseUp( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Virtual;
    Begin
      If GDragged = Self
      Then Begin
        UpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        GDragged    := Nil;
        EndMove( GDistance);
      End;
    End;

    procedure   TCustomDetail.DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer); // Virtual;
    Begin
      If GDragged = Self
      Then Begin
        UpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        UpdateMove( GDistance);
      End;
    End;

    Procedure   TCustomDetail.SendDblClick( Sender: TObject); // Virtual;
    Begin
      If Owner Is TCustomModule
      Then TCustomModule( Owner).SendDblClick( Sender)
      Else If Owner Is TDrawSheet
      Then TDrawSheet( Owner).SendDblClick( Sender);
    End;

    Procedure   TCustomDetail.SendRightClick( Sender: TObject); // Virtual;
    Begin
      If Owner Is TCustomModule
      Then TCustomModule( Owner).SendRightClick( Sender)
      Else If Owner Is TDrawSheet
      Then TDrawSheet( Owner).SendRightClick( Sender);
    End;

    Function    TCustomDetail.CenterPoint: TPoint;
    Begin
      With Result Do
      Begin
        x := Width  Div 2;
        y := Height Div 2;
      End;
    End;

    Procedure   TCustomDetail.InvalidateRectangle( aRect: TRect); // Virtual;
    Begin
      If Owner Is TCustomModule
      Then
        TCustomModule( Owner).
          InvalidateRectangle( MoveRect( aRect, Point( Left, Top)))
      Else If Owner Is TDrawSheet
      Then
        TDrawSheet( Owner).
          InvalidateRectangleDetail( MoveRect( aRect, Point( Left, Top)));
    End;

    Procedure   TCustomDetail.SetSizeAndPos( R: TRect); // Virtual;
    Var
      W      : Integer;
      H      : Integer;
      Old    : TRect;
    Begin
      Old := ParentRect;
      W   := R.Right  - R.Left;
      H   := R.Bottom - R.Top;
      Snap( R.TopLeft, FSnap);
      If
        ( FLeft   <> R.Left) Or
        ( FTop    <> R.Top ) Or
        ( FWidth  <> W     ) Or
        ( FHeight <> H     )
      Then Begin
        FLeft   := R.Left;
        FTop    := R.Top;
        FWidth  := W;
        FHeight := H;
        If Owner Is TCustomModule
        Then TCustomModule( Owner).InvalidateRectangle( Old)
        Else If Owner Is TDrawSheet
        Then TDrawSheet( Owner).InvalidateRectangleDetail( Old);
        Invalidate;
      End;
    End;

    Procedure   TCustomDetail.SetPen( aValue: TPen);
    Begin
      FPen.Assign( aValue);
      Invalidate;
    End;

    Procedure   TCustomDetail.SetBrush( aValue: TBrush);
    Begin
      FBrush.Assign( aValue);
      Invalidate;
    End;

    Procedure   TCustomDetail.SetLeft( aValue: Integer);
    Var
      R : TRect;
    Begin
      If FLeft <> aValue
      Then Begin
        R := ParentRect;
        R.Left := aValue;
        R.Right := R.Left + Width;
        SetSizeAndPos( R);
      End;
    End;

    Procedure   TCustomDetail.SetTop( aValue: Integer);
    Var
      R : TRect;
    Begin
      If FTop <> aValue
      Then Begin
        R := ParentRect;
        R.Top := aValue;
        R.Bottom := R.Top + Height;
        SetSizeAndPos( R);
      End;
    End;

    Procedure   TCustomDetail.SetWidth( aValue: Integer);
    Var
      R : TRect;
    Begin
      If FWidth <> aValue
      Then Begin
        R := ParentRect;
        R.Right := R.Left + aValue;
        SetSizeAndPos( R);
      End;
    End;

    Procedure   TCustomDetail.SetHeight( aValue: Integer);
    Var
      R : TRect;
    Begin
      If FHeight <> aValue
      Then Begin
        R := ParentRect;
        R.Bottom := R.Top + aValue;
        SetSizeAndPos( R);
      End;
    End;

    Function    TCustomDetail.GetParentRect: TRect;
    Begin
      Result := Rect( Left, Top, Left + Width, Top + Height);
    End;

    Function    TCustomDetail.GetClientRect: TRect;
    Begin
      Result := Rect( 0, 0, Width, Height);
    End;

    Function    TCustomDetail.GetScreenRect: TRect;
    Begin
      With Result Do
      Begin
        TopLeft     := ClientOrigin;
        BottomRight := AddPoints( ClientOrigin, Point( Width, Height));
      End;
    End;

    Function    TCustomDetail.GetClientOrigin: TPoint;
    Begin
//      Result := ClientToScreen( Point( Left, Top));
      Result := ClientToScreen( Point( 0, 0));
    End;

    Procedure   TCustomDetail.SetColor( aValue: TColor);
    Begin
      Brush.Color := aValue;
      Invalidate;
    End;

    Function    TCustomDetail.GetColor: TColor;
    Begin
      Result := Brush.Color;
    End;

{  Protected }

    Procedure   TCustomDetail.BeginMove( aPoint: TPoint); // Virtual;
    Begin
    End;

    Procedure   TCustomDetail.EndMove( aPoint: TPoint); // Virtual;
    Begin
    End;

    Procedure   TCustomDetail.UpdateMove( aPoint: TPoint); // Virtual
    Begin
    End;

    Procedure   TCustomDetail.Invalidate; // Virtual;
    Begin
      InvalidateRectangle( ClientRect);
    End;

    Procedure   TCustomDetail.PaintOn( aCanvas: TCanvas); // Virtual;
    Begin
      With aCanvas Do
      Begin
        Pen   := Self.Pen;
        Brush := Self.Brush;
        Rectangle( 0, 0, Width, Height);
      End;
    End;

    Function    TCustomDetail.ContainsPoint( aPoint: TPoint): Boolean; // Virtual;
    Begin
      Result := PtInRect( ClientRect, aPoint);
    End;

    Function    TCustomDetail.ClientToScreen( aPoint: TPoint): TPoint;
    Begin
      If ( Owner Is TCustomModule) And HasParent And ( Owner.HasParent)
      Then Result := AddPoints( TCustomModule( Owner).ClientToScreen( aPoint), Point( Left, Top))
      Else Result := AddPoints( aPoint, Point( Left, Top));
    End;

    Function    TCustomDetail.ScreenToClient( aPoint: TPoint): TPoint;
    Begin
      If ( Owner Is TCustomModule) And HasParent And Owner.HasParent
      Then Result := AddPoints( TCustomModule( Owner).ScreenToClient( aPoint), Point( - Left, - Top))
      Else Result := AddPoints( aPoint, Point( - Left, - Top));
    End;

    Procedure   TCustomDetail.SetParentComponent( aValue: TComponent); // Override;
    Begin
      Assert( aValue = Owner, 'Didn''t expect parent to be not the owner ...')
    End;

    procedure   TCustomDetail.AssignTo( aDest: TPersistent); // Override;
    Begin
      If aDest Is TCustomDetail
      Then Begin
        TCustomDetail( aDest).Left  := Left;
        TCustomDetail( aDest).Top   := Top;
        TCustomDetail( aDest).Pen   .Assign( Pen);
        TCustomDetail( aDest).Brush .Assign( Brush);
      End;
    End;

{  Public }

    Constructor TCustomDetail.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FPen          := TPen.Create;
      FBrush        := TBrush.Create;
      Brush.Color   := clBlue;
      Pen  .Color   := clBlack;
      FSnap         := Point( 1, 1);
      Width         := 25;
      Height        := 25;
      Name          := UniqueName;
      FHandlesMouse := True;
    End;

    Destructor  TCustomDetail.Destroy; // Override;
    Begin
      FPen  .Free;
      FBrush.Free;
      Inherited Destroy;
    End;

    Function    TCustomDetail.UniqueName: String; // Virtual;
    Var
      i : Integer;
      S : String;
    Begin
      Result := '';
      S      := Copy( ClassName, 2, Length( ClassName));
      With Owner Do
        For i := 0 To ComponentCount - 1 Do
          If FindComponent( Format( '%s%d', [ S, i])) = Nil
          Then Begin
            Result := Format( '%s%d', [ S, i]);
            Break;
          End;
    End;

    Procedure   TCustomDetail.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Virtual;
    Begin
      WriteLn( aFile, Indent( anInd, Format( '%s : ''%s''', [ ClassName, Name])));
    End;

    Function    TCustomDetail.GetParentComponent: TComponent; // Override;
    Begin
      Result := Owner;
    End;

    Function    TCustomDetail.HasParent: Boolean; // Override;
    Begin
      Result := True;
    End;

    Procedure   TCustomDetail.VisualMaintainance; // Virtual;
    Begin
      // Nothing, it's for children only.
    End;

{
  // a semi-abstract base type for details without mouse interaction
  // (such as labels).
  //
  TNoMouse = Class( TCustomDetail)
  Public
}

    Constructor TNoMouse.Create( anOwner: TComponent); // Override;
    Begin
      Inherited;
      FHandlesMouse := False;
    End;

{ ========
  TConnector = Class( TCustomDetail)
  Private
      FConnectionCount : Integer;
      FValue           : TValue;
      FConnectorType   : TConnectorType;
  Public
    Property    Value         : TValue         Read FValue         Write SetValue;
    Property    Connected     : Boolean        Read GetConnected;
    Property    ConnectorType : TConnectorType Read FConnectorType Write SetConnectorType;
  Private
}

    Procedure   TConnector.SetValue( aValue: TValue);
    Begin
      If FValue <> aValue
      Then Begin
        FValue := aValue;
        ValueChanged( FValue);
      End;
    End;

    Procedure   TConnector.SetConnectorType( aValue: TConnectorType);
    Begin
      If aValue <> FConnectorType
      Then FConnectorType := aValue;
    End;

    Function    TConnector.GetConnected: Boolean;
    Begin
      Result := FConnectionCount > 0;
    End;

    Procedure   TConnector.EndMove( aPoint: TPoint); // Override;
    Var
      aDst : TConnector;
    Begin
      aDst := SearchDestination;
      If Assigned( aDst) And aDst.CanAccept( Self)
      Then Connect( aDst);
    End;

    Procedure   TConnector.UpdateMove( aPoint: TPoint); // Override;
    Var
      aDC    : HDC;
      aPen   : HPen;
      OldObj : HGDIOBJ;
      aDst   : TConnector;
      aColor : TColor;
    Begin
      aDst := SearchDestination;
      If Assigned( aDst) And aDst.CanAccept( Self)
      Then aColor := clGreen
      Else aColor := clWhite;
      aDC := GetDC( 0);

      If Owner Is TCustomModule
      Then
        With TCustomModule( Owner) Do
          If Owner Is TDrawSheet
            Then
              With TDrawSheet( Owner) Do
                With ClientOrigin Do
                  IntersectClipRect( aDC, X, Y, X + ClientWidth, Y + ClientHeight);
      Try
        SetRop2( aDC, R2_XORPEN);
        aPen   := CreatePen( PS_SOLID, 0, ColorToRGB( aColor));
        OldObj := SelectObject( aDC, aPen);
        Try
          With ClientToScreen( GAnchor) Do
          Begin
            MoveToEx( aDC, x, y, Nil);
            LineTo( aDC, x + aPoint.x, y + aPoint.y);
          End;
        Finally
          SelectObject( aDC, OldObj);
          DeleteObject( aPen);
        End;
      Finally
        ReleaseDC( 0, aDC);
      End;
    End;

    Function    TConnector.GetBitmap: TBitmap; // Virtual;
    Begin
      Result := BmRoundGray;
    End;

    Procedure   TConnector.PaintOn( aCanvas: TCanvas); // Override;

      Procedure SetColors( aColor: TColor);
      Begin
        With aCanvas Do
        Begin
          Brush.Color := aColor;
          Pen  .Color := aColor;
        End;
      End;

    Begin
      With aCanvas Do
      Begin
        Draw( 0, 0, GetBitMap);
        If FConnectionCount > 0
        Then Begin
          Case ConnectorType Of
            ctSymmetric  : SetColors( clRed);
            ctAsymmetric : SetColors( clBlue);
            ctLogic      : SetColors( clYellow);
            ctNeuron     : SetColors( clGray);
          End;
          Ellipse( 3, 3, 8, 8); // Draw the plug, when connected.
        End;
      End;
    End;

    Procedure   TConnector.Connect( aConnector: TConnector); // Virtual;
    Begin
      If
        ( Owner Is TCustomModule) And
        ( aConnector <> Self    )
      Then
        With TCustomModule( Owner) Do
          Connect( Self, aConnector);
    End;

    Function    TConnector.CanAccept( aConnector: TConnector): Boolean; // Virtual;
    Begin
      Result := aConnector Is TConnector;
    End;

    Function    TConnector.SearchDestination: TConnector;
    Var
      aPoint: TPoint;
    Begin
      aPoint := AddPoints( GAnchor, GDistance);
      If ContainsPoint( aPoint)
      Then Result := Self
      Else
        If Owner Is TCustomModule
        Then
          Result :=
            TCustomModule( Owner).
              SearchDestination( Self, AddPoints( aPoint, Point( Left, Top)))
        Else
          Result := Nil;
    End;

    Procedure   TConnector.ValueChanged( Var aValue: TValue); // Virtual;
    Begin
      // Nothing here, it's for the children.
    End;

    procedure   TConnector.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If aDest Is TConnector
      Then TConnector( aDest).ConnectorType := ConnectorType;
    End;

{   Public }

    Constructor TConnector.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Width         := 10;
      Height        := 10;
      ConnectorType := ctAsymmetric;
    End;

    Procedure   TConnector.IncConnectionCount;
    Begin
      Inc( FconnectionCount);
      If FConnectionCount = 1
      Then Invalidate;
    End;

    Procedure   TConnector.DecConnectionCount;
    Begin
      If FConnectionCount > 0
      Then Begin
        Dec( FconnectionCount);
        If FConnectionCount = 0
        Then Invalidate;
      End;
    End;

{ ========
  TInput = Class( TConnector)
  Protected
}

    Function    TInput.GetBitmap: TBitmap; // Override;
    Begin
      Case ConnectorType Of
        ctSymmetric  : Result := BmRoundRed;
        ctAsymmetric : Result := BmRoundBlue;
        ctLogic      : Result := BmRoundYellow;
        ctNeuron     : Result := BmRoundGray;
        Else           Result := Inherited GetBitmap;
      End;
    End;

    Function    TInput.CanAccept( aConnector: TConnector): Boolean; // Override;
    Begin
      CanAccept := aConnector Is TOutput;
    End;


{ ========
  TOutput = Class( TConnector)
  Protected
}

    Function    TOutput.GetBitmap: TBitmap; // Override;
    Begin
      Case ConnectorType Of
        ctSymmetric  : Result := BmSquareRed;
        ctAsymmetric : Result := BmSquareBlue;
        ctLogic      : Result := BmSquareYellow;
        ctNeuron     : Result := BmSquareGray;
        Else           Result := Inherited GetBitmap;
      End;
    End;

    Function    TOutput.CanAccept( aConnector: TConnector): Boolean; // Override;
    Begin
      Result := aConnector Is TInput;
    End;

{
  TTextLabel = Class( TNoMouse)
  Private
      FText        : String;
      FFont        : TFont;
      FAutoSize    : Boolean;
      FParentColor : Boolean;
  Public
    Property    Text        : String  Read FText        Write SetText;
    Property    TextColor   : TColor  Read GetTextColor Write SetTextColor;
    Property    Font        : TFont   Read FFont        Write SetFont;
    Property    AutoSize    : Boolean Read FAutoSize    Write SetAutoSize;
    Property    ParentColor : Boolean Read FParentColor Write SetParentColor;
  Private
}

    Procedure   TTextLabel.AdjustBounds;
    var
      aDC       : HDC;
      aSaveFont : HFont;
      aTextSize : TSize;
    begin
      aDC := GetDC( 0);
      aSaveFont := SelectObject( aDC, Font.Handle);
      GetTextExtentPoint32( aDC, PChar( FText), Length( FText), aTextSize);
      SelectObject( aDC, aSaveFont);
      ReleaseDC( 0, aDC);
      SetSizeAndPos(
        Rect(
          Left,
          Top,
          Left + aTextSize.cx + ( GetSystemMetrics( SM_CXBORDER) * 2),
          Top  + aTextSize.cy + ( GetSystemMetrics( SM_CYBORDER) * 1)
        )
      );
    End;

    Procedure   TTextLabel.SetText( Const aValue: String);
    Begin
      If FText <> aValue
      Then Begin
        FText := aValue;
        If FAutoSize
        Then AdjustBounds;
        InvalidateRectangle( InflateRectangle( ClientRect, - 1));
      End;
    End;

    Procedure   TTextLabel.SetTextColor( aValue: TColor);
    Begin
      If TextColor <> aValue
      Then Begin
        Font.Color := aValue;
        InvalidateRectangle( InflateRectangle( ClientRect, - 1));
      End;
    End;

    Function    TTextLabel.GetTextColor: TColor;
    Begin
      Result := Font.Color;
    End;

    Procedure   TTextLabel.SetFont( aValue: TFont);
    Begin
      Font.Assign( aValue);
      If FAutoSize
      Then AdjustBounds
      Else InvalidateRectangle( InflateRectangle( ClientRect, - 1));
    End;

    Procedure   TTextLabel.SetAutoSize( aValue: Boolean);
    Begin
      If FAutoSize <> aValue
      Then Begin
        FAutoSize := aValue;
        If FAutoSize
        Then AdjustBounds;
      End;
    End;

    Procedure   TTextLabel.SetParentColor( aValue: Boolean);
    Begin
      If aValue <> FParentColor
      Then Begin
        FParentColor := aValue;
        Invalidate;
      End;
    End;

{  Protected }

    Procedure   TTextLabel.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      Pen.Color := Brush.Color;
      Inherited PaintOn( aCanvas);
      With aCanvas Do
      Begin
        If Not FParentColor
        Then Brush := Self.Brush;
        Font := Self.Font;
        TextOut( 0, 0, FText);
      End;
    End;

    procedure   TTextLabel.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If aDest Is TTextLabel
      Then Begin
        TTextLabel( aDest).Font.Assign( Font);
        TTextLabel( aDest).Text        := Text;
        TTextLabel( aDest).AutoSize    := AutoSize;
        TTextLabel( aDest).ParentColor := ParentColor;
      End;
    End;

{  Public }

    Constructor TTextLabel.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FFont := TFont.Create;
      With FFont Do
      Begin
        Name   := 'Arial';
        Size   := 7;
        Style  := [];
      End;
      Color       := ModuleColor;
      TextColor   := clBlack;
      Width       := 90;
      Height      := 15;
      AutoSize    := False;
      ParentColor := True;
    End;

    Destructor  TTextLabel.Destroy; // Override;
    Begin
      FFont.Free;
      Inherited Destroy;
    End;

{ ========
  TViewer = Class( TCustomDetail)
  Private
      FText : String;
      FFont : TFont;
  Public
    Property    Text        : String Read FText          Write SetText;
    Property    BorderColor : TColor Read GetBorderColor Write SetBorderColor;
    Property    TextColor   : TColor Read GetTextColor   Write SetTextColor;
    Property    Font        : TFont  Read FFont          Write SetFont;
  Private
}

    Procedure   TViewer.SetText( Const aValue: String);
    Begin
      If FText <> aValue
      Then Begin
        FText      := aValue;
        FUpdateReq := True;
      End;
    End;

    Procedure   TViewer.SetBorderColor( aValue: TColor);
    Begin
      If BorderColor <> aValue
      Then Begin
        Pen.Color := aValue;
        Invalidate;
      End;
    End;

    Function    TViewer.GetBorderColor: TColor;
    Begin
      Result := Pen.Color;
    End;

    Procedure   TViewer.SetTextColor( aValue: TColor);
    Begin
      If TextColor <> aValue
      Then Begin
        Font.Color := aValue;
        InvalidateRectangle( InflateRectangle( ClientRect, - 1));
      End;
    End;

    Function    TViewer.GetTextColor: TColor;
    Begin
      Result := Font.Color;
    End;

    Procedure   TViewer.SetFont( aValue: TFont);
    Begin
      Font.Assign( aValue);
      InvalidateRectangle( InflateRectangle( ClientRect, - 1));
    End;

{  Protected }

    Procedure   TViewer.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      With aCanvas Do
      Begin
        Pen   := Self.Pen;
        Brush := Self.Brush;
        Font  := Self.Font;
        Rectangle( 0, 0, Width, Height);
        TextOut( 4, 1, FText);
      End;
    End;

    procedure   TViewer.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If aDest Is TViewer
      Then Begin
        TViewer( aDest).Text        := Text;
        TViewer( aDest).BorderColor := BorderColor;
        TViewer( aDest).TextColor   := TextColor;
      End;
    End;

{  Public }

    Constructor TViewer.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FFont       := TFont.Create;
      Width       := 35;
      Height      := 15;
      Color       := ViewerColor;
      BorderColor := clWhite;
      TextColor   := clWhite;
      With FFont Do
      Begin
        Size  := 8;
        Style := [];
      End;
      FUpdateReq := True;
    End;

    Destructor  TViewer.Destroy; // Override;
    Begin
      FFont.Free;
      Inherited Destroy;
    End;

    Procedure   TViewer.VisualMaintainance; // Override;
    Begin
      If FUpdateReq
      Then Begin
        InvalidateRectangle( InflateRectangle( ClientRect, - 2));
        FUpdateReq := False;
      End;
    End;
    

{ ========
  TKnob = Class( TCustomDetail)
  Private
      FValue         : TValue;
      FMinValue      : TValue;
      FMaxValue      : TValue;
      FViewer        : TViewer;
      FOnValueChange : TOnValueValueChange;
      FOldP1         : TPoint;
      FOldP2         : TPoint;
      FKnobMode      : TKnobMode;
      FRepeatCount   : Integer;
  Public
    Property    Value        : TValue              Read FValue         Write SetValue;
    Property    MinValue     : TValue              Read FMinValue      Write SetMinValue;
    Property    MaxValue     : TValue              Read FMaxValue      Write SetMaxValue;
    Property    Viewer       : TViewer             Read FViewer        Write SetViewer;
    Property    KnobMode     : TKnobMode           Read FKnobMode;
    Property    OnValueChange: TOnValueValueChange Read FOnValueChange Write SetOnValueChange;
  Private
}

    Procedure   TKnob.DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Override;
    Begin
      If ( GDragged = Nil) And ( Button = mbLeft) And Not ( ssDouble In Shift)
      Then Begin
        GDragged     := Self;
        GLastDragged := GDragged;
        GAnchor      := Point( X, Y);
        GDistance    := Point( X, Y);
        BeginMove ( GAnchor  );
        UpdateMove( GDistance);
      End
      Else Inherited;
    End;

    Procedure   TKnob.DoTimerFired( aSender: TObject);
    Begin
      Inc( FRepeatCount);
      If FRepeatCount < 50
      Then GRepeatTimer.Interval := RepeatPause
      Else GRepeatTimer.Interval := FastRepeatPause;
      If FKnobMode In [ kmInc,kmDec]
      Then Begin
        Try
          Case FKnobMode Of
            kmInc  : Value := Value + 1;
            kmDec  : Value := Value - 1;
          End;
        Except
          GRepeatTimer.Enabled := False;
          FRepeatCount := 0;
          Raise;
        End;
      End;
    End;

    Procedure   TKnob.UpdatePeers;
    Begin
      If Assigned( FViewer)
      Then FViewer.Text := IntToStr( Value);
      If Assigned( FOnValueChange)
      Then FOnValueChange( Self, FValue);
    End;

    Procedure   TKnob.SetValue( aValue: TValue);
    Var
      R1  : TRect;
      R2  : TRect;
      P1  : TPoint;
      P2  : TPoint;
      Ang : Integer;
    Begin
      If aValue <> FValue
      Then Begin
        Ang := ValuesToAngle( MinValue, aValue, MaxValue);
        P1  := PolarToPoint( CenterPoint, Ang, 1);
        P2  := PolarToPoint( CenterPoint, Ang, Width Div 2 - 1);
        If P1.x > P2.x
        Then Exchange( P1.x, P2.x);
        If P1.y > P2.y
        Then Exchange( P1.y, P2.y);

        Clip( aValue, FMinValue, FMaxValue);
        FValue := aValue;

        UpdatePeers;

        If
          ( P1.x  <> FOldP1.x) Or
          ( P1.y  <> FOldP1.y) Or
          ( P2.x  <> FOldP2.x) Or
          ( P2.y  <> FOldP2.y)
        Then Begin
          R1 := InflateRectangle( Rect( P1.x    , P1.y    , P2.x    , P2.y)    , 1);
          R2 := InflateRectangle( Rect( FOldP1.x, FOldP1.y, FOldP2.x, FOldP2.y), 1);
          InvalidateRectangle( R2);
          InvalidateRectangle( R1);
          FOldP1 := P1;
          FOldP2 := P2;
        End;
      End;
    End;

    Procedure   TKnob.SetMinValue( aValue: TValue);
    Begin
      If aValue <> FMinValue
      Then Begin
        FMinValue := aValue;
        If FValue < FMinValue
        Then FValue := FMinValue;
        Invalidate;
      End;
    End;

    Procedure   TKnob.SetMaxValue( aValue: TValue);
    Begin
      If aValue <> FMaxValue
      Then Begin
        FMaxValue := aValue;
        If FValue > FMaxValue
        Then FValue := FMaxValue;
        Invalidate;
      End;
    End;

    Procedure   TKnob.SetViewer( aValue: TViewer);
    Begin
      If aValue <> FViewer
      Then Begin
        FViewer := aValue;
        If Assigned( FViewer)
        Then Invalidate;
      End;
    End;

    Procedure   TKnob.SetOnValueChange( aValue: TOnValueValueChange);
    Begin
      FOnValueChange := aValue;
      If Assigned( FOnValueChange)
      Then Invalidate;
    End;

{  Protected }

    Procedure   TKnob.BeginMove( aPoint: TPoint); // Override;
    Const
      CRange   = 8;
      CRadius  = 25;
    Var
      R1 : TRect;
      R2 : TRect;
      P  : TPoint;
    Begin
      R1 := Rect( 0             , Height - CRange, CRange, Height);
      R2 := Rect( Width - CRange, Height - CRange, Width , Height);

      If PtInRect( R1, aPoint)
      Then FKnobMode := kmDec
      Else If PtInRect( R2, aPoint)
      Then FKnobMode := kmInc
      Else FKnobMode := kmTurn;

      If FKnobMode = kmTurn
      Then Begin
        P          := PolarToPoint( CenterPoint, ValuesToAngle( MinValue, Value, MaxValue), CRadius);
        With ClientToScreen( P) Do SetCursorPos( X, Y);
        GDistance  := SubPoints( CenterPoint, P);
        GAnchor    := CenterPoint;
      End
      Else Begin
        FRepeatCount := 0;
        With GRepeatTimer Do
        Begin
          OnTimer  := DoTimerFired;
          Interval := InitRepeatPause;
          Enabled  := True;
        End;
      End;
    End;

    Procedure   TKnob.EndMove( aPoint: TPoint); // Override;
    Var
      aValue : TValue;
    Begin
      With GRepeatTimer Do
      Begin
        OnTimer := Nil;
        Enabled := False;
      End;
      Case FKnobMode Of
        kmInc  : aValue := Value + 1;
        kmDec  : aValue := Value - 1;
        Else     aValue := DistanceToValue( MinValue, MaxValue, aPoint);
      End;
      Clip( aValue, MinValue, MaxValue);
      Value     := aValue;
      FKnobMode := kmNone;
    End;

    Procedure   TKnob.UpdateMove( aPoint: TPoint); // Override;
    Begin
      Case FKnobMode Of
        kmTurn : Value := DistanceToValue( MinValue, MaxValue, aPoint);
        kmInc  : {Nothing};
        kmDec  : {Nothing};
      End;
    End;

    Function    TKnob.GetBitmap: TBitmap; // Virtual;
    Begin
      Result := BmKnob;
    End;

    Procedure   TKnob.PaintOn( aCanvas: TCanvas); // Override;
    Var
      S : TPoint;
      E : TPoint;
      A : Integer;
    Begin
      With aCanvas Do
      Begin
        Pen := Self.Pen;
        Draw( 0, 0, GetBitmap);
        A := ValuesToAngle( MinValue, Value, MaxValue);
        S := PolarToPoint( CenterPoint, A, 1);
        E := PolarToPoint( CenterPoint, A, Width Div 2 - 1);
        With aCanvas Do
        Begin
          MoveTo( S.x, S.y);
          LineTo( E.x, E.y);
        End;
      End;
    End;

    procedure   TKnob.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If aDest Is TKnob
      Then Begin
        TKnob( aDest).Value    := Value;
        TKnob( aDest).MinValue := MinValue;
        TKnob( aDest).MaxValue := MaxValue;
      End;
    End;

{  Public }

    Constructor TKnob.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Width     :=   29;
      Height    :=   29;
      Value     :=  500;
      MinValue  :=    0;
      MaxValue  := 1000;
      Pen.Color := clWhite;
    End;

    Procedure   TKnob.Invalidate; // Override;
    Begin
      Inherited Invalidate;
      UpdatePeers;
    End;


{ ========
  TKnobMedium = Class( TKnob)
  Protected
}

    Function    TKnobMedium.GetBitmap: TBitmap; // Override;
    Begin
      Result := BmKnobMedium;
    End;

//  Public

    Constructor TKnobMedium.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Width  := 21;
      Height := 21;
    End;


{ ========
  TKnobSmall = Class( TKnob)
  Protected
}

    Function    TKnobSmall.GetBitmap: TBitmap; // Override;
    Begin
      Result := BmKnobSmall;
    End;

//  Public

    Constructor TKnobSmall.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Width  := 15;
      Height := 15;
    End;


{ ========
  TExpKnob = Class( TKnob)
  Protected
}

    Procedure   TExpKnob.EndMove( aPoint: TPoint); // Override;
    Begin
      Value := DistanceToExpValue( MinValue, MaxValue, aPoint);
    End;

    Procedure   TExpKnob.UpdateMove( aPoint: TPoint); // Override;
    Begin
      Value := DistanceToExpValue( MinValue, MaxValue, aPoint);
    End;


{ ========
  TPushButton = Class( TCustomDetail)
  Private
      FDown          : Boolean;
      FOldDown       : Boolean;
      FBitmap        : TBitmap;
      FOnValueChange : TOnBooleanValueChange;
      FGroupId       : Integer;
      FAllowAllUp    : Boolean;
      FOnClick       : TNotifyEvent;
  Protected
    Property    Bitmap: TBitmap Read FBitmap Write SetBitmap;
  Public
    Property    Down         : Boolean               Read FDown          Write SetDown;
    Property    GroupId      : Integer               Read FGroupId       Write SetGroupId;
    Property    AllowAllUp   : Boolean               Read FAllowAllUp    Write SetAllowAllUp;
    Property    OnValueChange: TOnBooleanValueChange Read FOnValueChange Write SetOnValueChange;
    Property    OnClick      : TNotifyEvent          Read FOnClick       Write FOnClick;
  Private
}

    Procedure   TPushButton.SelectUnique;
    Var
      i    : Integer;
      aBut : TPushButton;
    Begin
      If Owner Is TCustomDetail
      Then
        With TCustomDetail( Owner) Do
          For i := 0 To ComponentCount - 1 Do
            If ( Components[ i] Is TPushButton) And ( Components[ i] <> Self)
            Then Begin
              aBut := TPushButton( Components[ i]);
              With aBut Do
                If ( GroupId = Self.GroupId) And FDown
                Then Begin
                  FDown := False;
                  If Assigned( FOnValueChange)
                  Then FOnValueChange( aBut, False);
                End;
            End;
    End;

    Procedure   TPushButton.SetBitmap( aValue: TBitmap);
    Begin
      With FBitmap Do
      Begin
        Assign( aValue);
        TransparentMode := tmAuto;
        Transparent     := True;
      End;
      Invalidate;
    End;

    Procedure   TPushButton.SetDown( aValue: Boolean);
    Begin
      If FDown <> aValue
      Then Begin
        If ( GroupId > 0) And Not aValue And Not AllowAllup
        Then Exit;
        FDown := aValue;
        If Down And ( GroupId > 0)
        Then SelectUnique;
        If Assigned( FOnValueChange)
        Then FOnValueChange( Self, Down);
        Invalidate;
      End;
    End;

    Procedure   TPushButton.SetOnValueChange( aValue: TOnBooleanValueChange);
    Begin
      FOnValueChange := aValue;
      If Assigned( FOnValueChange) And ( GDragged <> Self)
      Then FOnValueChange( Self, Down);
    End;

    Procedure   TPushButton.SetGroupId( aValue: Integer);
    Begin
      If FGroupId <> aValue
      Then Begin
        FGroupId := aValue;
        If FGroupId > 0
        Then Begin
          If Down
          Then SelectUnique
          Else
            If Not AllowAllUp
            Then Down := True;
        End;
      End;
    End;

    Procedure   TPushButton.SetAllowAllUp( aValue: Boolean);
    Begin
      If FAllowAllUp <> aValue
      Then Begin
        FAllowAllUp := aValue;
        If ( GroupId > 0) And Not FAllowAllUp And Not Down
        Then Down := True;
      End;
    End;

{  Protected }

    Procedure   TPushButton.BeginMove( aPoint: TPoint); // Override;
    Begin
      GAnchor  := Point( 0, 0);
      FOldDown := Down;
      Down     := Not Down;
    End;

    Procedure   TPushButton.EndMove( aPoint: TPoint); // Override;
    Begin
      If ContainsPoint( aPoint)
      Then Begin
        Down := Not FOldDown;
        If Assigned( FOnClick)
        Then FOnClick( Self);
      End
      Else Down := FOldDown;
    End;

    Procedure   TPushButton.UpdateMove( aPoint: TPoint); // Override;
    Begin
      Down := ContainsPoint( aPoint) Xor FOldDown;
    End;

    Procedure   TPushButton.PaintOn( aCanvas: TCanvas); // Override;
    Var
      R1 : TRect;
      R2 : TRect;
      C1 : TColor;
      C2 : TColor;
    Begin
      R1 := InflateRectangle( ClientRect, - 1);
      With R1 Do
      Begin
        Dec( Right );
        Dec( Bottom);
      End;
      R2 := InflateRectangle( ClientRect, - 2);
      With aCanvas Do
      Begin
        Self.Pen.Color := ModuleColor;
        If Down
        Then Self.Brush.Color := $00D0D0A0
        Else Self.Brush.Color := clSilver;
        Brush := Self.Brush;
        Pen   := Self.Pen;
        Rectangle( 0, 0, Width, Height);
        If Down
        Then Begin
          C1 := clGray;
          C2 := clWhite;
          R2 := MoveRect( R2, Point( 1, 1));
        End
        Else Begin
          C1 := clWhite;
          C2 := clGray;
        End;
        Self.Pen.Color := C1;
        Pen   := Self.Pen;
        With R1 Do
        Begin
          MoveTo( Left , Bottom);
          LineTo( Left , Top   );
          LineTo( Right, Top   );
          Self.Pen.Color := C2;
          Pen := Self.Pen;
          MoveTo( Left , Bottom);
          LineTo( Right, Bottom);
          LineTo( Right, Top   );
        End;
        With R2 Do
        Begin
          MoveWindowOrg( Handle, Left, Top);
          IntersectClipRect( Handle, 0, 0, Right - 4, Bottom - 4);
        End;
        Draw( 0, 0, FBitmap);
      End;
    End;

    procedure   TPushButton.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If aDest Is TPushButton
      Then Begin
        TPushButton( aDest).Down   := Down;
        TPushButton( aDest).Bitmap .Assign( Bitmap);
      End;
    End;

{  Public }

    Constructor TPushButton.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FBitmap := TBitmap.Create;
      FBitmap.TransparentMode := tmAuto;
      Brush.Color := ModuleColor;
      Pen  .Color := ModuleColor;
    End;

    Destructor  TPushButton.Destroy; // Override;
    Begin
      FBitmap.Free;
      Inherited Destroy;
    End;

    Procedure   TPushButton.LoadBitmap( Const aFileName: String);
    Var
      Bm : TBitmap;
    Begin
      Bm := CreateBitmapFromFile( aFileName);
      Try
        With Bitmap Do
        Begin
          Assign( Bm);
          TransparentMode := tmAuto;
          Transparent     := True;
        End;
      Finally
        Bm.Free;
      End;
    End;

{ ========
  TIllustration = Class( TNoMouse)
  Private
      FBitmap      : TBitmap;
      FTransparent : Boolean;
      FAutoSize    : Boolean;
  Public
    Property    Bitmap      : TBitmap Read FBitmap      Write SetBitmap;
    Property    Transparent : Boolean Read FTransparent Write SetTransparent;
    Property    AutoSize    : Boolean Read FAutoSize    Write SetAutoSize;
  Private
}

    Procedure   TIllustration.AdjustBounds;
    Var
      R : TRect;
    Begin
      R := ParentRect;
      With R Do
      Begin
        Bottom := Top  + FBitmap.Height;
        Right  := Left + FBitmap.Width;
      End;
      SetSizeAndPos( R);
      Invalidate;
    End;

    Procedure   TIllustration.SetBitmap( aValue: TBitmap);
    Begin
      With FBitmap Do
      Begin
        Assign( aValue);
        TransparentMode := tmAuto;
        Transparent     := FTransparent;
      End;
      If FAutoSize
      Then AdjustBounds
      Else Invalidate;
    End;

    Procedure   TIllustration.SetTransparent( aValue: Boolean);
    Begin
      If FTransparent <> aValue
      Then Begin
        FTransparent := aValue;
        FBitmap.Transparent := FTransparent;
        Invalidate;
      End;
    End;

    Procedure   TIllustration.SetAutoSize( aValue: Boolean);
    Begin
      If FAutoSize <> aValue
      Then Begin
        FAutoSize := aValue;
        If FAutoSize
        Then AdjustBounds;
      End;
    End;

{  Protected }

    Procedure   TIllustration.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      With aCanvas Do
        Draw( 0, 0, FBitmap);
    End;

    procedure   TIllustration.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If aDest Is TIllustration
      Then Begin
        TIllustration( aDest).Bitmap.Assign( Bitmap);
        TIllustration( aDest).Transparent := Transparent;
        TIllustration( aDest).AutoSize    := AutoSize;
      End;
    End;

{  Public }

    Constructor TIllustration.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Width  := 20;
      Height := 20;
      FBitmap := TBitmap.Create;
      FTransparent := True;
      Transparent  := True;
      AutoSize     := False;
    End;

    Destructor  TIllustration.Destroy; // Override;
    Begin
      FBitmap.Free;
      Inherited Destroy;
    End;

    Procedure   TIllustration.LoadBitmap( Const aFileName: String);
    Var
      Bm : TBitmap;
    Begin
      Bm := CreateBitmapFromFile( aFileName);
      Try
        Bitmap := Bm;
      Finally
        Bm.Free;
      End;
    End;


{ ========
Const

  MaxPoints = 8192;

Type

  TPointArray = Array[ 0 .. MaxPoints - 1] Of TPoint;
  PPointArray = ^TPointArray;

  TGraphSheet = Class( TNoMouse)
  Private
      FData      : PPointArray;
      FEltCount  : Integer;
      FXScale    : Integer;
      FYScale    : Integer;
      FDrawColor : TColor;
  Public
    Property    Data   [ anIndex: Integer]: TPoint  Read GetData    Write SetData;
    Property    XData  [ anIndex: Integer]: Integer Read GetXData   Write SetXData;
    Property    YData  [ anIndex: Integer]: Integer Read GetYData   Write SetYData;
    Property    XScaled[ anIndex: Integer]: Integer Read GetXScaled;
    Property    YScaled[ anIndex: Integer]: Integer Read GetYScaled;
    Property    EltCount                  : Integer Read FEltCount  Write SetEltCount;
    Property    XScale                    : Integer Read FXScale    Write SetXScale;
    Property    YScale                    : Integer Read FYScale    Write SetYScale;
    Property    DrawColor                 : TColor  Read FDrawColor Write SetDrawColor;
  Private
}

    Procedure   TGraphSheet.CheckBounds( anIndex: Integer);
    Begin
      If ( anIndex < 0) Or ( anIndex >= EltCount)
      Then Raise EBounds.CreateFmt( 'GraphSheet::Index %d out of bounds [%d,%d]', [ anIndex, 0, EltCount - 1])
    End;

    Function    TGraphSheet.GetData( anIndex: Integer): TPoint;
    Begin
      CheckBounds( anIndex);
      Result := FData^[ anIndex];
    End;

    Procedure   TGraphSheet.SetData( anIndex: Integer; aValue: TPoint);
    Var
      R : TRect;
    Begin
      CheckBounds( anIndex);
      If ( FData^[ anIndex].x <> aValue.x) Or ( FData^[ anIndex].y <> aValue.y)
      Then Begin
        R :=
          Rect(
            XScaled[ anIndex],
            YScaled[ anIndex],
            XScaled[ anIndex] + 1,
            YScaled[ anIndex] + 1
          );
        FData^[ anIndex] := aValue;
        InvalidateRectangle( R);

        InvalidateRectangle(
          Rect(
            XScaled[ anIndex],
            YScaled[ anIndex],
            XScaled[ anIndex] + 1,
            YScaled[ anIndex] + 1
          )
        );
      
      End;
    End;

    Function    TGraphSheet.GetXData( anIndex: Integer): Integer;
    Begin
      Result := Data[ anIndex].x;
    End;

    Procedure   TGraphSheet.SetXData( anIndex: Integer; aValue: Integer);
    Begin
      With Data[ anIndex] Do
        x := aValue;
    End;

    Function    TGraphSheet.GetYData( anIndex: Integer): Integer;
    Begin
      Result := Data[ anIndex].y;
    End;

    Procedure   TGraphSheet.SetYData( anIndex: Integer; aValue: Integer);
    Begin
      With Data[ anIndex] Do
        y := aValue;
    End;

    Function    TGraphSheet.GetXScaled( anIndex: Integer): Integer;
    Begin
      Result := Round( 1 + ( XData[ anIndex] / 1000) * ( XScale / 1000)* ( Width - 2));
    End;

    Function    TGraphSheet.GetYScaled( anIndex: Integer): Integer;
    Begin
      Result := Round( 1 - (( YData[ anIndex] - 1000) / 2000) * ( YScale / 1000) * ( Height - 2));
    End;

    Procedure   TGraphSheet.SetEltCount ( aValue: Integer);
    Var
      NewData : PPointArray;
      i       : Integer;
    Begin
      If FEltCount <> aValue
      Then Begin
        NewData := Nil;
        If aValue > 0
        Then Begin
          GetMem( NewData, aValue * SizeOf( TPoint));
          For i := 0 To aValue - 1 Do
            NewData^[ i] := Point( 0, 0);
        End;
        If Assigned( FData) And ( FEltCount > 0)
        Then Begin
          If Assigned( NewData)
          Then Move( FData^, NewData^, Minimum( FEltCount, aValue));
          FreeMem( FData, FEltCount * SizeOf( TPoint));
        End;
        FEltCount := aValue;
        FData     := NewData;
      End;
    End;

    Procedure   TGraphSheet.SetXScale( aValue: Integer);
    Begin
      If FXScale <> aValue
      Then Begin
        FXScale := aValue;
        Invalidate;
      End;
    End;

    Procedure   TGraphSheet.SetYScale( aValue: Integer);
    Begin
      If FYScale <> aValue
      Then Begin
        FYScale := aValue;
        Invalidate;
      End;
    End;

    Procedure   TGraphSheet.SetDrawColor( aValue: TColor);
    Begin
      If FDrawColor <> aValue
      Then Begin
        FDrawColor := aValue;
        Invalidate;
      End;
    End;

//   Protected

    Procedure   TGraphSheet.PaintOn( aCanvas: TCanvas); // Override;
    Const
      VM = 8;
      HM = 8;
    Var
      i     : Integer;
      PrevX : Integer;
      PrevY : Integer;
      NewX  : Integer;
      NewY  : Integer;
    Begin
      Inherited PaintOn( aCanvas);
      With aCanvas Do
      Begin
        For i := 1 To VM - 1 Do
        Begin
          MoveTo( i * Width Div VM, 0     );
          LineTo( i * Width Div VM, Height);
        End;
        For i := 1 To HM - 1 Do
        Begin
          MoveTo( 0    , i * Height Div HM);
          LineTo( Width, i * Height Div HM);
        End;
        PrevX := MaxInt;
        PrevY := MaxInt;
        If EltCount > 0
        Then Begin
          For i := 0 To EltCount - 1 Do
          Begin
            NewX := XScaled[ i];
            NewY := YScaled[ i];
            If ( NewX <> PrevX) Or ( NewY <> PrevY)
            Then Begin
              Pixels[ NewX, NewY] := DrawColor;
              PrevX := NewX;
              PrevY := NewY;
            End;
          End;
        End;
      End;
    End;

    procedure   TGraphSheet.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If aDest Is TGraphSheet
      Then Begin
        TGraphSheet( aDest).EltCount  := EltCount;
        TGraphSheet( aDest).XScale    := XScale;
        TGraphSheet( aDest).YScale    := YScale;
        TGraphSheet( aDest).DrawColor := DrawColor;
      End;
    End;

//   Public

    Constructor TGraphSheet.Create( anOwner: TComponent); // Override;
    Var
      i : Integer;
    Begin
      Inherited Create( anOwner);
      EltCount    := 128;
      XScale      := 1000;
      YScale      := 1000;
      DrawColor   := GraphForePen;
      Brush.Color := GraphBack;
      Pen  .Color := GraphBackPen;
      For i := 0 To EltCount - 1 Do
      Begin
        XData[ i] := Random( 1001);
        YData[ i] := Random( 1001);
      End;
    End;

    Destructor  TGraphSheet.Destroy; // Override;
    Begin
      EltCount := 0;
      Inherited Destroy;
    End;

    Procedure   TGraphSheet.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Override;
    Var
      i : Integer;
    Begin
      Inherited Dump( aFile, anInd, aDepth);
      If aDepth > 0
      Then Begin
        WriteLn( aFile, Indent( anInd, 'Data'));
        For i := 0 To EltCount - 1 Do
          WriteLn(
            aFile,
            Indent( anInd + 1,
              Format(
                '( %.5d, %.5d) ( %.5d, %.5d)',
                [
                  XData  [ i],
                  YData  [ i],
                  XScaled[ i],
                  YScaled[ i]
                ]
              )
            )
          );
        WriteLn( aFile, Indent( anInd, 'End data'));
      End;
    End;

    Procedure   TGraphSheet.Clear;
    Begin
      FUpdateReq := True;
    End;

    Procedure   TGraphSheet.VisualMaintainance; // Override;
    Begin
      If FUpdateReq
      Then Begin
        Invalidate;
        FUpdateReq := False;
      End;
    End;

{ ========
  TIndicator = Class( TNoMouse)
  Private
      FActive   : Boolean;
  Public
    Property    Active   : Boolean Read FActive   Write SetActive;
  Private
}

    Procedure   TIndicator.SetActive( aValue: Boolean);
    Begin
      If FActive <> aValue
      Then Begin
        FActive    := aValue;
        FUpdateReq := True;
      End;
    End;

{  Protected }

    Function    TIndicator.GetBitmap: TBitmap; // Virtual;
    Begin
      If Factive
      Then Result := BmLedOn
      Else Result := BmLedOff;
    End;

    Procedure   TIndicator.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      aCanvas.Draw( 0, 0, GetBitMap);
    End;

    procedure   TIndicator.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
    End;

{  Public }

    Constructor TIndicator.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Width      := 9;
      Height     := 4;
      FUpdateReq := True;
    End;

    Procedure   TIndicator.VisualMaintainance; // Override;
    Begin
      If FUpdateReq
      Then Begin
        Invalidate;
        FUpdateReq := False;
      End;
    End;

{ ========
  TCustomModule = Class( TCustomPanel)
  Private
      FSelected   : Boolean;
      FInCreation : Boolean;
      FTextLabel  : TTextLabel;
  Public
    Property    Selected  : Boolean    Read FSelected  Write SetSelected;
    Property    TextLabel : TTextLabel Read FTextLabel Write FTextLabel;
  Published
    Property    Title     : String     Read GetTitle   Write SetTitle;
  Private
}

    Procedure   TCustomModule.DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Override;
    Var
      aDetail : TCustomDetail;
    Begin
      If GDragged = Nil
      Then Begin
        aDetail := SearchDetail( Point( X, Y));
        If Assigned( aDetail)
        Then Begin
          If ssDouble In Shift
          Then Begin
            If aDetail = Self
            Then SendDblClick( aDetail)
            Else
              With aDetail Do
                DoMouseDown( Sender, Button, Shift, X - Left, Y - Top);
          End
          Else If Button = mbRight
          Then Begin
            If aDetail = Self
            Then SendRightClick( aDetail)
            Else
              With aDetail Do
                DoMouseDown( Sender, Button, Shift, X - Left, Y - Top);
          End
          Else If Button = mbLeft
          Then Begin
            If ( aDetail <> Self) And aDetail.FHandlesMouse
            Then
              With aDetail Do
                DoMouseDown( Sender, Button, Shift, X - Left, Y - Top)
            Else Begin
              GDragged     := Self;
              GLastDragged := Self;
              GAnchor      := Point( X, Y);
              GDistance    := Point( 0, 0);
              HandleSelection ( ssCtrl In Shift);
              BeginMove       ( GAnchor  );
              HandleUpdateMove( GDistance);
            End;
          End;
        End;
      End;
    End;

    Procedure   TCustomModule.DoMouseUp( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Override;
    Begin
      If GDragged = Self
      Then Begin
        Inherited;
        ScrollSelfInView;
        If FInCreation
        Then Begin
          FInCreation := False;
          Invalidate;
        End;
        Restack;
      End
      Else Inherited;
    End;

    procedure   TCustomModule.DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer); // Override;
    Begin
      If GDragged = Self
      Then Begin
        HandleUpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        HandleUpdateMove( GDistance);
      End;
    End;

    Procedure   TCustomModule.SetSelected( aValue: Boolean);
    Begin
      If aValue <> FSelected
      Then Begin
        FSelected := aValue;
        HandleSelectionChange( FSelected);
      End;
    End;

    Procedure   TCustomModule.HandleSelection( CtrlKeyActive: Boolean);
    Begin
      If CtrlKeyActive
      Then Selected := Not Selected
      Else Begin
        If Not Selected
        Then UnselectPeers;
        Selected := True;
      End;
      BringToFront;
    End;

    Function    TCustomModule.GetTitle: String;
    Begin
      Result := TextLabel.Text;
    End;

    Procedure   TCustomModule.SetTitle( Const aValue: String);
    Begin
      TextLabel.Text := aValue;
    End;

    Procedure   TCustomModule.HandleUpdateMove( aPoint: TPoint);
    Begin
      If Owner Is TDrawSheet
      Then TDrawSheet( Owner).HandleModuleUpdateMove( Self, aPoint)
      Else UpdateMove( aPoint);
    End;

    Procedure   TCustomModule.UnselectPeers;
    Begin
      If Owner Is TDrawSheet
      Then TDrawSheet( Owner).UnselectAllModulesBut( Self);
    End;

    Procedure   TCustomModule.BringToFront;
    Begin
      If Owner Is TDrawSheet
      Then TDrawSheet( Owner).BringModuleToFront( Self)
    End;

    Procedure   TCustomModule.Disconnect( aConnector: TConnector); // Virtual;
    Begin
      If Owner Is TDrawSheet
      Then TDrawSheet( Owner).Disconnect( aConnector)
    End;

    Procedure   TCustomModule.MarkConnections;
    Var
      i : Integer;
    Begin
      If Owner Is TDrawSheet
      Then Begin
        For i := 0 To ComponentCount - 1 Do
          If Components[ i] Is TConnector
          Then TDrawSheet( Owner).MarkConnections( TConnector( Components[ i]));
      End;
    End;

{  Protected }

    Procedure   TCustomModule.Notification( aComponent: TComponent; anOperation: TOperation); // Override;
    Begin
      If ( aComponent Is TConnector) And ( anOperation = opRemove)
      Then Disconnect( TConnector( aComponent));
    End;

    Procedure   TCustomModule.HandleSelectionChange( aValue: Boolean); // Virtual;
    Begin
      If Not ( csDestroying In ComponentState)
      Then
        With FTextLabel Do
          Case aValue Of
            False : Color := ModuleColor;
            True  : Color := clWhite;
          End;
    End;

    Procedure   TCustomModule.Restack;
    {}
      Function DoSort( Item1, Item2: Pointer): Integer;
      {}
        Function SortModules( F1, F2: TCustomModule): Integer;
        Begin
          If F1.Left < F2.Left
          Then Result := -1
          Else If F1.Left > F2.Left
          Then Result := 1
          Else If F1.Top < F2.Top
          Then Result := -1
          Else If F1.Top > F2.Top
          Then Result := 1
          Else Result := 0;
        End;
      {}
      Begin
        Result := SortModules( TCustomModule( Item1), TCustomModule( Item2));
      End;
    {}
    Var
      LocModules : TList;
      i          : Integer;
      CTop       : Integer;
      CLeft      : Integer;
    Begin
      If HasParent
      Then Begin
        LocModules := TList.Create;
        Try
          If Owner Is TDrawSheet
          Then
            With TDrawSheet( Owner) Do
            Begin
              For i := ModuleCount - 1 Downto 0 Do
              Begin
                With Module[ i] Do
                  If Left < 0
                  Then Left := 0;
                LocModules.Add( Module[ i]);
              End;
            End;
          With LocModules Do
          Begin
            Sort( @ DoSort);
            CLeft := 0;
            CTop  := 0;
            For i := 0 To Count - 1 Do
            Begin
              With TCustomModule( Items[ i]) Do
              Begin
                If Left > CLeft
                Then Begin
                  CLeft := Left;
                  CTop  := 0;
                End;
                If Top < CTop
                Then Top := CTop;
                CTop := Top + Height;
              End;
            End;
          End;
        Finally
          LocModules.Free;
        End;
      End;
    End;

    Procedure   TCustomModule.EndMove( aPoint: TPoint); // Override;
    Begin
      If Owner Is TDrawSheet
      Then TDrawSheet( Owner).HandleModuleMoved( aPoint)
      Else Move( aPoint);
    End;

    Procedure   TCustomModule.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      If Not FInCreation
      Then Begin
        With aCanvas Do
        Begin
          Pen   := Self.Pen;
          Brush := Self.Brush;
          Rectangle( 0, 0, Width, Height);
          Pen.Color := clWhite;
          MoveTo( 0        , Height - 2);
          LineTo( 0        , 0         );
          LineTo( Width - 1, 0         );
          Pen.Color := clBlack;
        End;
        PaintDetails( aCanvas);
      End;
    End;

    Procedure   TCustomModule.PaintDetails( aCanvas: TCanvas); // Virtual;
    Var
      i   : Integer;
      Sdc : Integer;
      Res : Bool;
    Const
      Recur  : Boolean = False;
    Begin
      Assert( Not Recur, 'Oops ... recursion ...');
      Recur := True;
      With aCanvas Do
      Begin
        For i := 0 To ComponentCount - 1 Do
          If Components[ i] Is TCustomDetail
          Then
            With TCustomDetail( Components[ i]) Do
            Begin
              Sdc := SaveDC( Handle);
              Assert( Sdc <> 0, 'No more save DC ...');
              MoveWindowOrg( Handle, Left, Top);
              IntersectClipRect( Handle, 0, 0, Width, Height);
              Refresh;
              Try
                If RectanglesOverlap( ClientRect, ClipRect)
                Then PaintOn( aCanvas);
              Finally
                Res := RestoreDC( Handle, - 1);
                Assert( Res, 'No more restore DC ...');
              End;
            End;
      End;
      Recur := False;
    End;

    Procedure   TCustomModule.UpdateMove( aPoint: TPoint); // Override;
    Var
      aDC   : HDC;
      aRect : TRect;
    Begin
      aRect := ClientRect;
      aDC := GetDC( 0);
      If Owner Is TDrawSheet
      Then
        With TDrawSheet( Owner) Do
        Begin
          With ClientOrigin Do
            IntersectClipRect( aDC, X, Y, X + ClientWidth, Y + ClientHeight);
          With aRect Do
          Begin
            Left   := ClientOrigin.x + Self.Left + aPoint.x;
            Top    := ClientOrigin.y + Self.Top  + aPoint.y;
            Right  := Left + Self.Width;
            Bottom := Top  + Self.Height;
          End;
        End;
      Try
        DrawFocusRect( aDC, aRect);
      Finally
        ReleaseDC( 0, aDC);
      End;
    End;

    Procedure   TCustomModule.Move( aPoint: TPoint); // Virtual;
    Begin
      SetSizeAndPos( MoveRect( ParentRect, aPoint));
    End;

    Procedure   TCustomModule.CreateDetails; // Virtual;
    Begin
      FTextLabel := TTextLabel( CreateDetail( 2, 2, '', TTextLabel));
      With FTextLabel Do
      Begin
        AutoSize      := False;
        Width         := 55;
        Height        := 12;
        Text          := Copy( Self.ClassName, 2, Length( ClassName));
        ParentColor   := False;
        Color         := ModuleColor;
        FHandlesMouse := True;
      End;
    End;

    Function    TCustomModule.CreateDetail( aLeft, aTop: Integer; Const aName: String; aDetailClass: TDetailClass): TCustomDetail;
    Begin
      Result := aDetailClass.Create( Self);

      If ( Result Is TConnector) And ( aName = '')
      Then Raise EInvalidName.Create( 'Connectors must have a non-empty name !');

      With Result Do
      Begin
        Left := aLeft;
        Top  := aTop;
        Name := aName;
      End;
    End;

    Function    TCustomModule.SearchDestination( aSource: TConnector; aPoint: TPoint): TConnector;
    Begin
      Result := FindDestination( aSource, aPoint);
      If ( Result = Nil) And ( Owner Is TDrawSheet)
      Then
        Result :=
          TDrawSheet( Owner).
            FindDestination(
              aSource,
              AddPoints(
                Point(
                  Left,
                  Top
                ),
                aPoint
              )
            );
    End;

    Function    TCustomModule.SearchDetail( aPoint: TPoint): TCustomDetail;
    Var
      i : Integer;
      C : TCustomDetail;
    Begin
      Result := Nil;
      If ContainsPoint( aPoint)
      Then Begin
        Result := Self;
        For i := ComponentCount - 1 DownTo 0 Do
          If Components[ i] Is TCustomDetail
          Then Begin
            C := TCustomDetail( Components[ i]);
            With C Do
              If ContainsPoint( AddPoints( aPoint, Point( - Left, - Top)))
              Then Begin
                Result := C;
                Break;
              End;
          End;
      End;
    End;

    Function    TCustomModule.FindDestination( aSource: TConnector; aPoint: TPoint): TConnector;
    Var
      i : Integer;
      C : TConnector;
    Begin
      Result := Nil;
      If ContainsPoint( aPoint)
      Then
        For i := ComponentCount - 1 Downto 0 Do
          If ( Components[ i] Is TConnector) And ( Components[ i] <> aSource)
          Then Begin
            C := TConnector( Components[ i]);
            With C Do
              If ContainsPoint( AddPoints( aPoint, Point( - Left, - Top)))
              Then Begin
                Result := C;
                Break;
              End;
          End;
    End;

    Procedure   TCustomModule.Connect( aSource, aDestination: TConnector);  // Virtual;
    Begin
      If Owner Is TDrawSheet
      Then TDrawSheet( Owner).Connect( aSource, aDestination);
    End;

    Function    TCustomModule.ClientToScreen( aPoint: TPoint): TPoint; // Override;
    Begin
      Result := aPoint;
      If ( Owner Is TDrawSheet) And Owner.HasParent
      Then Result := AddPoints( TDrawSheet( Owner).ClientToScreen( aPoint), Point( Left, Top));
    End;

    Function    TCustomModule.ScreenToClient( aPoint: TPoint): TPoint; // Override;
    Begin
      Result := aPoint;
      If ( Owner Is TDrawSheet) And Owner.HasPArent
      Then Result := AddPoints( TDrawSheet( Owner).ScreenToClient( aPoint), Point( - Left, - Top));
    End;

    Procedure   TCustomModule.ScrollSelfInView;
    Begin
      If Owner Is TDrawSheet
      Then TDrawSheet( Owner).ScrollModuleInView( Self);
    End;

    procedure   TCustomModule.AssignTo( aDest: TPersistent); // Override;
    Var
      i : Integer;
    Begin
      Inherited AssignTo( aDest);
      If aDest Is TCustomModule
      Then Begin
        For i := 0 To ComponentCount - 1 Do
          TCustomModule( aDest).Components[ i].Assign( Components[ i]);
        TCustomModule( aDest).Selected := Selected;
      End;
    End;

{  Public }

    Constructor TCustomModule.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FInCreation := GAutoCreation;
      FSnap       := Point( 250, 50);
      Left        := 0;
      Top         := 0;
      Width       := 250;
      Height      := 50;
      Color       := ModuleColor;
      Brush.Color := Color;
      Pen  .Color := clBlack;
      CreateDetails;
      FSelected   := True;
      Selected    := False;
    End;

    Destructor  TCustomModule.Destroy; // Override;
    Var
      i : Integer;
    Begin
      For i := 0 To ComponentCount - 1 Do
        If Components[ i] Is TCustomDetail
        Then Notification( Components[ i], opRemove);
      Inherited Destroy;
    End;

    Procedure   TCustomModule.Execute; // Virtual;
    Begin
      // Nothing here, it's for the children.
    End;

    Procedure   TCustomModule.VisualMaintainance; // Override;
    Var
      i : Integer;
    Begin
      For i := 0 To ComponentCount - 1 Do
        If Components[ i] Is TCustomDetail
        Then
          With TCustomDetail( Components[ i]) Do
            If FUpdateReq
            Then VisualMaintainance;
    End;
    
    Procedure   TCustomModule.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Override;
    Var
      i : Integer;
    Begin
      Inherited Dump( aFile, anInd, aDepth);
      If aDepth > 0
      Then
        For i := 0 To ComponentCount - 1 Do
          If Components[ i] Is TCustomDetail
          Then TCustomDetail( Components[ i]).Dump( aFile, anInd + 1, aDepth - 1);
    End;

{ ========
  TModule2 = Class( TCustomModule)
  Public
}

    Constructor TModule2.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Height := Height * 2;
    End;

{ ========
  TModule4 = Class( TCustomModule)
  Public
}

    Constructor TModule4.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Height := Height * 4;
    End;

{ ========
  TModule8 = Class( TCustomModule)
  Public
}

    Constructor TModule8.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Height := Height * 8;
    End;

{ ========
  TMidiModule = Class( TCustomModule)
  Protected
}

    Procedure   TMidiModule.AcceptMidiShort( aShortMidi: TShortMidi); // Virtual;
    Begin
    End;

    Procedure   TMidiModule.AcceptMidiLong( aLongMidi: TLongMidi); // Virtual;
    Begin
    End;

{  Public }

    Procedure   TMidiModule.MidiOutShort( aShortMidi: TShortMidi); // Virtual;
    Begin
      If Owner Is TDrawSheet
      Then TDrawSheet( Owner).PerformMidiOutShort( Self, aShortMidi);
    End;

    Procedure   TMidiModule.MidiOutLong( Const aLongMidi: TLongMidi); // Virtual;
    Begin
      If Owner Is TDrawSheet
      Then TDrawSheet( Owner).PerformMidiOutLong( Self, aLongMidi);
    End;


{ ========
  TConnection = Class
  Private
      FSource      : TConnector;
      FDestination : TConnector;
      FMarked      : Boolean;
      FSelected    : Boolean;
  Public
    Property    Source      : TConnector Read FSource;
    Property    Destination : TConnector Read FDestination;
    Property    Selected    : Boolean    Read FSelected     Write SetSelected;
  Private
}

    Procedure   TConnection.SetSelected( aValue: Boolean);
    Begin
      FSelected := aValue;
    End;

{ Public }

    Constructor TConnection.Create( aSource, aDestination: TConnector);
    Begin
      Inherited Create;
      FSource      := aSource;
      FDestination := aDestination;
    End;

    Destructor  TConnection.Destroy; // Override;
    Begin
      If FSource Is TInput
      Then Source.Value := 0;
      If FDestination Is TInput
      Then Destination.Value := 0;
      Inherited Destroy;
    End;

{ ========
  TConnections = Class( TList)
  Private
      FOwner : TDrawSheet;
  Public
    Property    ConnectionCount : Integer Read GetConnectionCount;
    Property    Connection [ anIndex: Integer]: TConnection Read GetConnection;
    Property    Owner: TDrawSheet Read FOwner;
  Private
}

    Procedure   TConnections.Disconnect( aConnector: TConnector);
    Var
      i : Integer;
    Begin
      For i := ConnectionCount - 1 Downto 0 Do
        With Connection[ i] Do
          If ( Source = aConnector) Or ( Destination = aConnector)
          Then Begin
            Source     .DecConnectionCount;
            Destination.DecConnectionCount;
            Connection[ i].Free;
            Delete( i);
            FOwner.ConnectionLost( Source, Destination)
          End;
      Pack;
    End;

    Function    TConnections.GetConnectionCount: Integer;
    Begin
      Result := Count;
    End;

    Function    TConnections.GetConnection( anIndex: Integer): TConnection;
    Begin
      Result := TConnection( Items[ anIndex]);
    End;

{  Public }

    Constructor TConnections.Create( anOwner: TDrawSheet);
    Begin
      Inherited Create;
      FOwner := anOwner;
    End;

    Function    TConnections.Connect( aSource, aDestination: TConnector): Boolean;
    Var
      i : Integer;
    Begin
      i := FindConnection( aSource, aDestination);
      If i < 0
      Then i := FindConnection( aDestination, aSource);
      If i >= 0
      Then Begin
        aSource     .DecConnectionCount;
        aDestination.DecConnectionCount;
        Connection[ i].Free;
        Delete( i);
        Pack;
        Result := False;
      End
      Else Begin
        If aSource Is TInput
        Then Add( TConnection.Create( aDestination, aSource))
        Else Add( TConnection.Create( aSource, aDestination));
        aSource     .IncConnectionCount;
        aDestination.IncConnectionCount;
        Result := True;
      End;
    End;

    Function    TConnections.FindConnection( aSource, aDestination: TConnector): Integer;
    Var
      i : Integer;
    Begin
      Result := - 1;
      For i := ConnectionCount - 1 DownTo 0 Do
        With Connection[ i] Do
          If ( aSource = Source) And ( aDestination = Destination)
          Then Begin
            Result := i;
            Break;
          End;
    End;

    Procedure   TConnections.Execute; // Virtual;
    Var
      i : Integer;
    Begin
      // Clear all input values.
      For i := 0 To ConnectionCount - 1 Do
        With Connection[ i] Do
          If Source Is TInput
          Then Source.FValue := 0
          Else If Destination Is TInput
          Then Destination.FValue := 0;
      // Transfer outputvalues to input values.
      For i := 0 To ConnectionCount - 1 Do
        With Connection[ i] Do
          If ( Source Is TOutput) And ( Destination Is TInput)
          Then
            With Destination Do
              FValue := FValue + Source.FValue
          Else If ( Source Is TInput) And ( Destination Is TOutput)
          Then
            With Source Do
              FValue := FValue + Destination.FValue
          Else If ( Destination Is TInput) And ( Source Is TInput)
          Then
            With Source Do
              FValue := FValue + Destination.FValue;
    End;

{ ========
  TDrawSheet = Class( TGraphicControl)
  Private
      FModules           : TList;
      FConnections       : TConnections;
      FSlot              : TSlot;
      FCanvas            : TCanvas;
      FPen               : TPen;
      FSlotActive        : Boolean;
      FDoDrawConnections : Boolean;
      FConnectionFixups  : TStringList;
  Public
    Property    Pen        : TPen    Read FPen        Write SetPen;
    Property    SlotActive : Boolean Read FSlotActive Write SetSlotActive;
    Property    Module     [ anIndex: Integer]: TCustomModule  Read GetModule;
    Property    Connection [ anIndex: Integer]: TConnection    Read GetConnection;
    Property    Source     [ anIndex: Integer]: TConnector     Read GetSource;
    Property    Destination[ anIndex: Integer]: TConnector     Read GetDestination;
  Private
}

    procedure   TDrawSheet.WMPaint( Var aMessage: TWMPaint); // Message WM_PAINT;
    Begin
      If aMessage.DC <> 0
      then begin
        FCanvas.Handle := aMessage.DC;
        try
          Paint;
        finally
          FCanvas.Handle := 0;
        end;
      end;
    End;

    Procedure   TDrawSheet.DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Var
      aDetail : TCustomDetail;
    Begin
      If GDragged = Nil
      Then Begin
        aDetail := SearchDetail( Point( X, Y));
        If Assigned( aDetail)
        Then
          With aDetail Do
            DoMouseDown( Sender, Button, Shift, X - Left, Y - Top)
        Else If Button = mbLeft
        Then Begin
          GDragged     := Self;
          GLastDragged := Self;
          GAnchor      := Point( X, Y);
          GDistance    := Point( 0, 0);
          BeginMove ( GAnchor  );
          UpdateMove( GDistance);
        End;
      End;
    End;

    Procedure   TDrawSheet.DoMouseUp( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Var
      aPoint: TPoint;
    Begin
      If GDragged = Self
      Then Begin
        UpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        GDragged := Nil;
        EndMove( GDistance);
      End
      Else
        If GDragged Is TCustomDetail
        Then
          With TCustomDetail( GDragged) Do
          Begin
            aPoint := ScreenToClient( Self.ClientToScreen( Point( X, Y)));
            With aPoint Do
              DoMouseUp( Sender, Button, Shift, X, Y);
          End;
    End;

    procedure   TDrawSheet.DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer);
    Var
      aPoint : TPoint;
    Begin
      If GDragged = Self
      Then Begin
        UpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        UpdateMove( GDistance);
      End
      Else
        If GDragged Is TCustomDetail
        Then
          With TCustomDetail( GDragged) Do
          Begin
            aPoint := ScreenToClient( Self.ClientToScreen( Point( X, Y)));
            With aPoint Do
              DoMouseMove( Sender, Shift, X, Y);
          End;
    End;

    Procedure   TDrawSheet.SendDblClick( Sender: TObject);
    Begin
      If Assigned( FSlot)
      Then FSlot.SendDblClick( Sender);
    End;

    Procedure   TDrawSheet.SendRightClick( Sender: TObject);
    Begin
      If Assigned( FSlot)
      Then FSlot.SendRightClick( Sender);
    End;

    Procedure   TDrawSheet.SetPen( aValue: TPen);
    Begin
      FPen.Assign( aValue);
      Invalidate;
    End;

    Procedure   TDrawSheet.SetSlotActive( aValue: Boolean);
    Begin
      If ( aValue <> FSlotActive) And Assigned( FSlot)
      Then Begin
        FSlot.Active := aValue;
        FSlotActive := aValue;
      End;
    End;

    Function    TDrawSheet.GetModule( anIndex: Integer): TCustomModule;
    Begin
      If ( anIndex >= 0) And ( anIndex < ModuleCount)
      Then Result := TCustomModule( FModules[ anIndex])
      Else Result := Nil;
    End;

    Function    TDrawSheet.GetConnection( anIndex: Integer): TConnection;
    Begin
      If ( anIndex >= 0) And ( anIndex < ConnectionCount)
      Then Result := FConnections[ anIndex]
      Else Result := Nil;
    End;

    Function    TDrawSheet.GetSource( anIndex: Integer): TConnector;
    Begin
      If ( anIndex >= 0) And ( anIndex < ConnectionCount)
      Then Result := Connection[ anIndex].Source
      Else Result := Nil;
    End;

    Function    TDrawSheet.GetDestination( anIndex: Integer): TConnector;
    Begin
      If ( anIndex >= 0) And ( anIndex < ConnectionCount)
      Then Result := Connection[ anIndex].Destination
      Else Result := Nil;
    End;

    Procedure   TDrawSheet.PerformCountChange;
    Begin
      If Assigned( FSlot)
      Then FSlot.PerformCountChange;
    End;

    Procedure   TDrawSheet.HandleModuleUpdateMove( aModule: TCustomModule; aPoint: TPoint);
    Var
      i : Integer;
    Begin
      For i := 0 to ModuleCount - 1 Do
        If Module[ i].Selected
        Then Module[ i].UpdateMove( aPoint);
    End;

    Procedure   TDrawSheet.HandleModuleMoved( aPoint: TPoint);
    Var
      i : Integer;
    Begin
      If IsNearNullPoint( aPoint)
      Then Exit;
      FDoDrawConnections := False;
      For i := 0 to ModuleCount - 1 Do
        If Module[ i].Selected
        Then Module[ i].MarkConnections;
      InvalidateMarkedConnections;
      For i := 0 to ModuleCount - 1 Do
        If Module[ i].Selected
        Then Begin
          With Module[ i] Do
          Begin
            MarkConnections;
            Module[ i].Move( aPoint);
          End;
        End;
      FDoDrawConnections := True;
      InvalidateMarkedConnections;
    End;

    Procedure   TDrawSheet.UnselectAllModulesBut( aModule: TCustomModule);
    Var
      i : Integer;
    Begin
      For i := 0 to ModuleCount - 1 Do
        If Module[ i].Selected And ( Module[ i] <> aModule)
        Then Module[ i].Selected := False;
    End;

    Procedure   TDrawSheet.Disconnect( aConnector: TConnector);
    Begin
      FConnections.Disconnect( aConnector);
      Invalidate;
    End;

    Procedure   TDrawSheet.InvalidateMarkedConnections;
    Var
      i  : Integer;
      R  : TRect;
      DR : PRect;
    Begin
      If Not ( csCreating In ControlState)
      Then
        If HasParent // And Parent.HandleAllocated
        Then Begin
          For i := 0 To ConnectionCount - 1 Do
          Begin
            With Connection[ i] Do
              If FMarked
              Then Begin
                UnionRect( R, Source.ScreenRect, Destination.ScreenRect);
                FMarked := False;
              End;
            R := MoveRect( R, ScreenToClient( Point( Left, Top)));
            GetMem( DR, SizeOf( TRect));
            DR^ := R;
            PostMessage( Parent.Handle, UM_PAINT, Integer( DR), 0);
//            InvalidateRect( Parent.Handle, @ R, True);
          End;
        End;
    End;

    Procedure   TDrawSheet.MarkConnections( aConnector: TConnector);
    Var
      i : Integer;
    Begin
      For i := 0 To ConnectionCount - 1 Do
        With Connection[ i] Do
          If ( Source = aConnector) Or ( Destination = aConnector)
          Then FMarked := True;
    End;

    Procedure   TDrawSheet.ClearConnectionFixups;
    Begin
      If Assigned( FConnectionFixUps)
      Then FConnectionFixups.Clear
      Else FConnectionFixups := TStringList.Create;
    End;

    Procedure   TDrawSheet.AddConnectionFixup( Const S: String);
    Begin
      FConnectionFixups.Add( S);
    End;

    Procedure   TDrawSheet.FixupConnections;
    Var
      i            : Integer;
      HadException : Boolean;
    Begin
      Try
        HadException := False;
        With FConnectionFixups Do
          For i := 0 To Count - 1 Do
            Try
              CreateConnectionFromFixup( Strings[ i]);
            Except
              On E:ECorruptPatch Do
                If Not HadException
                Then Begin
                  HadException := True;
                  ShowMessage( E.Message);
                End;
            End;
      Finally
        FConnectionFixups.Free;
        FConnectionFixups := Nil;
      End;
    End;

    Procedure   TDrawSheet.CreateConnectionFromFixup( Const S: String);
    Const
      Nums : Set Of Char = [ '0' .. '9'];
      Ascs : Set Of Char = [ '0' .. '9', 'A' .. 'Z', 'a' .. 'z'];
    Var
      p        : Integer;
      i        : Integer;
      SrcModId : Integer;
      SrcConId : Integer;
      DstModId : Integer;
      DstConId : Integer;
      SrcModNm : String;
      SrcConNm : String;
      DstModNm : String;
      DstConNm : String;
      SrcMod   : TCustomModule;
      SrcCon   : TConnector;
      DstMod   : TCustomModule;
      DstCon   : TConnector;
    Begin
      p := 1;
      i := 1;

      If ( Length( S) > 0) And ( S[ 1] In Nums)
      Then Begin
        // Do it the old way.

        While ( p <= Length( S)) And ( S[ p] In Nums) Do
          Inc( p);
        SrcModId := StrToInt( Copy( S, i, p - i));
        If ( p > Length( S)) Or ( S[ p] <> '.')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And ( S[ p] In Nums) Do
          Inc( p);
        SrcConId := StrToInt( Copy( S, i, p - i));
        If ( p > Length( S)) Or ( S[ p] <> ' ')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And ( S[ p] In Nums) Do
          Inc( p);
        DstModId := StrToInt( Copy( S, i, p - i));
        If ( p > Length( S)) Or ( S[ p] <> '.')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And ( S[ p] In Nums) Do
          Inc( p);
        DstConId := StrToInt( Copy( S, i, p - i));

        Try
          SrcMod := Components[ SrcModId] As TCustomModule;
          SrcCon := SrcMod.Components[ SrcConId] As TConnector;
          DstMod := Components[ DstModId] As TCustomModule;
          DstCon := DstMod.Components[ DstConId] As TConnector;
          Connect( SrcCon, DstCon);
        Except
          On Exception Do
            Raise
              ECorruptPatch.
                Create(
                  'The (old format) patch is corrupted or incompatible:'^M +
                  'can''t make all the connections, and some connections will probably be wrong.'
                );
        End;
      End

      Else If Length( S) > 0
      Then Begin
        // Do it the new way ....

        While ( p <= Length( S)) And ( S[ p] In Ascs) Do
          Inc( p);
        SrcModNm := Copy( S, i, p - i);
        If ( p > Length( S)) Or ( S[ p] <> '.')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And ( S[ p] In Ascs) Do
          Inc( p);
        SrcConNm := Copy( S, i, p - i);
        If ( p > Length( S)) Or ( S[ p] <> ' ')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And ( S[ p] In Ascs) Do
          Inc( p);
        DstModNm := Copy( S, i, p - i);
        If ( p > Length( S)) Or ( S[ p] <> '.')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And ( S[ p] In Ascs) Do
          Inc( p);
        DstConNm := Copy( S, i, p - i);

        Try
          SrcMod :=        FindComponent( SrcModNm) As TCustomModule;
          SrcCon := SrcMod.FindComponent( SrcConNm) As TConnector;
          DstMod :=        FindComponent( DstModNm) As TCustomModule;
          DstCon := DstMod.FindComponent( DstConNm) As TConnector;
          Connect( SrcCon, DstCon);
        Except
          On Exception Do
            Raise
              ECorruptPatch.
                Create(
                  'The (new format) patch is corrupted or incompatible:'^M +
                  'can''t make all the connections.'
                );
        End;

      End
      Else
        Raise
          ECorruptPatch.
            Create(
              'The patch is corrupted, incompatible or not a patch.'^M
            );
    End;

    Procedure   TDrawSheet.SyncModules;
    Var
      i : Integer;
    Begin
      FModules.Clear;
      For i := 0 To ComponentCount - 1 Do
        If Components[ i] Is TCustomModule
        Then Begin
          FModules.Add( Components[ i]);
          With TCustomModule( Components[ i]) Do
            Name := UniqueName;
        End;
    End;

    Procedure   TDrawSheet.PerformMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Begin
      If Assigned( FSlot)
      Then FSlot.PerformMidiOutShort( aSender, aShortMidi);
    End;

    Procedure   TDrawSheet.PerformMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
    Begin
      If Assigned( FSlot)
      Then FSlot.PerformMidiOutLong( aSender, aLongMidi);
    End;

{  Protected }

    Procedure   TDrawSheet.BeginMove( aPoint: TPoint); // Virtual;
    Begin
    End;

    Procedure   TDrawSheet.EndMove( aPoint: TPoint); // Virtual;
    Begin
      SelectModulesInRect(
        Rect(
          GAnchor.x,
          GAnchor.y,
          GAnchor.x + aPoint.x,
          GAnchor.y + aPoint.y
        )
      );
    End;

    Procedure   TDrawSheet.UpdateMove( aPoint: TPoint); // Virtual;
    Var
      aDC   : HDC;
      aRect : TRect;
    Begin
      aDC := GetDC( 0);
      With ClientOrigin Do
        IntersectClipRect( aDC, X, Y, X + ClientWidth, Y + ClientHeight);
      Try
        With aRect Do
        Begin
          Left   := Self.ClientOrigin.x + GAnchor.x;
          Top    := Self.ClientOrigin.y + GAnchor.y;
          Right  := Left + aPoint.x;
          Bottom := Top  + aPoint.y;
          If Left > Right
          Then Exchange( Left, Right);
          If Top > Bottom
          Then Exchange( Bottom, Top);
        End;
        DrawFocusRect( aDC, aRect);
      Finally
        ReleaseDC( 0, aDC);
      End;
    End;

    Procedure   TDrawSheet.Notification( aComponent: TComponent; anOperation: TOperation); // Override;
    Var
      i : Integer;
    Begin
      If aComponent Is TCustomModule
      Then Begin
        If anOperation = opInsert
        Then Begin
          i := IndexOfModule( TCustomModule( aComponent));
          If i < 0
          Then FModules.Add( aComponent);
        End
        Else If anOperation = opRemove
        Then Begin
          i := IndexOfModule( TCustomModule( aComponent));
          If i >= 0
          Then
            With FModules Do
            Begin
              Delete( i);
              Pack;
            End;
        End;
        PerformCountChange;
      End
      Else Inherited Notification( aComponent, anOperation);
    End;

    Procedure   TDrawSheet.BringModuleToFront( aModule: TCustomModule);
    Begin
      With FModules Do
        Exchange( IndexOfModule( aModule), Count - 1);
    End;

    Procedure   TDrawSheet.SelectModulesInRect( aRect: TRect);
    Var
      i : Integer;
    Begin
      With aRect Do
      Begin
        If Left > Right
        Then Exchange( Left, Right);
        If Top > Bottom
        Then Exchange( Bottom, Top);
      End;
      For i := 0 To ComponentCount - 1 Do
        If Components[ i] Is TCustomModule
        Then
          With TCustomModule( Components[ i]) Do
            Selected := RectanglesOverlap( aRect, Rect( Left, Top, Left + Width, Top + Height));
    End;

    Procedure   TDrawSheet.InvalidateRectangleDetail( aRect: TRect);
    Var
      R  : TRect;
      DR : PRect;
    Begin
      If Visible
      Then Begin
        If Assigned( Parent) // And Parent.HandleAllocated And Not ( csCreating In ControlState)
        Then Begin
          R := MoveRect( aRect, Point( Left, Top));
          GetMem( DR, SizeOf( TRect));
          DR^ := R;
          PostMessage( Parent.Handle, UM_PAINT, Integer( DR), 0);
//          InvalidateRect( Parent.Handle, @ R, True);
        End;
      End;
    End;

    Function    TDrawSheet.SearchDetail( aPoint: TPoint): TCustomDetail;
    Var
      i : Integer;
    Begin
      Result := Nil;
      If ContainsPoint( aPoint)
      Then Begin
        For i := ModuleCount - 1 DownTo 0 Do
          With Module[ i] Do
            If ContainsPoint( AddPoints( aPoint, Point( - Left, - Top)))
            Then Begin
              Result := Module[ i];
              Break;
            End;
      End;
    End;

    Function    TDrawSheet.ContainsPoint( aPoint: TPoint): Boolean; // Virtual;
    Begin
      Result := PtInRect( ClientRect, aPoint);
    End;

    Function    TDrawSheet.FindDestination( aSource: TConnector; aPoint: TPoint): TConnector;
    Var
      i : Integer;
    Begin
      Result := Nil;
      If ContainsPoint( aPoint)
      Then
        For i := ModuleCount - 1 DownTo 0 Do
        Begin
          Result :=
            Module[ i].
              FindDestination(
                aSource,
                AddPoints(
                  aPoint,
                  Point(
                    - Module[ i].Left,
                    - Module[ i].Top
                  )
                )
              );
          If Result <> Nil
          Then Break;
        End;
    End;

    Procedure   TDrawSheet.Connect( aSource, aDestination: TConnector); // Virtual;
    Var
      R  : TRect;
      DR : PRect;
    Begin
      If Assigned( aDestination)
      Then Begin
        If FConnections.Connect( aSource, aDestination)
        then ConnectionMade( aSource, aDestination)
        Else ConnectionLost( aSource, aDestination);
        If
          UnionRect( R, aSource.ScreenRect, aDestination.ScreenRect) And
          HasParent
        Then Begin
          R := MoveRect( R, ScreenToClient( Point( Left, Top)));
          GetMem( DR, SizeOf( TRect));
          DR^ := R;
          PostMessage( Parent.Handle, UM_PAINT, Integer( DR), 0);
//            InvalidateRect( Parent.Handle, @ R, True);
        End;
      End;
    End;

    Procedure   TDrawSheet.Paint; // Virtual;
    Begin
      PaintDetails( FCanvas);
      If FDoDrawConnections
      Then PaintConnections( FCanvas);
    End;

    Procedure   TDrawSheet.PaintDetails( aCanvas: TCanvas);
    Var
      i   : Integer;
      Sdc : Integer;
      Res : Bool;
    Const
      Recur  : Boolean = False;
    Begin
      Assert( Not Recur, 'Oops ... recursion ...');
      Recur := True;
      Try
        For i := 0 to ModuleCount - 1 Do
          With aCanvas, Module[ i] Do
          Begin
            Sdc := SaveDC( Handle);
            Assert( Sdc <> 0, 'No more save DC ...');
            MoveWindowOrg( Handle, Left, Top);
            IntersectClipRect( Handle, 0, 0, Width, Height);
            Refresh;
            Try
              If RectanglesOverlap( ClientRect, ClipRect)
              Then PaintOn( aCanvas);
            Finally
              Res := RestoreDC( Handle, - 1);
              Assert( Res, 'No more restore DC ...');
            End;
          End;
      Finally
        Recur := False;
      End;
    End;

    Procedure   TDrawSheet.PaintConnection( aCanvas: TCanvas; aConnection: TConnection);
    Var
      Points : TBezierPoints;
    Const
      Colors : Array[ TConnectorType] Of TColor = ( clRed, clBlue, clYellow, clGray);
    Begin
      With aConnection Do
      Begin
        Pen.Color := Colors[ Source.ConnectorType];
        With aCanvas Do
        Begin
          Pen := Self.Pen;
          With Source Do
            Points[ 0] := ClientToScreen( CenterPoint);
          Points[ 0] := ScreenToClient( Points[ 0]);
          With Destination Do
            Points[ 3] := ClientToScreen( CenterPoint);
          Points[ 3] := ScreenToClient( Points[ 3]);

          CalcBezierCurve( Points);
          If BezierIntersectsRectangle( Points, ClipRect)
          Then PolyBezier( Points);
        End;
      End;
    End;

    Procedure   TDrawSheet.PaintConnections( aCanvas: TCanvas);
    Var
      i : Integer;
    Begin
      aCanvas.Pen := Self.Pen;
      For i := 0 To ConnectionCount - 1 Do
        PaintConnection( aCanvas, Connection[ i]);
    End;

    Procedure   TDrawSheet.ScrollModuleInView( aModule: TCustomModule);
    Begin
      If Assigned( FSlot)
      Then FSlot.ScrollModuleInView( aModule);
    End;


    Procedure   TDrawSheet.ConnectionMade( aSource, aDestination: TConnector); // Virtual;
    Begin
      If Assigned( FSlot)
      Then FSlot.ConnectionMade( aSource, aDestination);
    End;

    Procedure   TDrawSheet.ConnectionLost( aSource, aDestination: TConnector);  //Virtual;
    Begin
      If Assigned( FSlot)
      Then FSlot.ConnectionLost( aSource, aDestination);
    End;

    ///////////////////////////////////////////////////////////////////////////
    //
    // Reader / Writer / Persistence - stuff
    //
    ///////////////////////////////////////////////////////////////////////////


    Procedure   TDrawSheet.GetChildren( aProc: TGetChildProc; Root: TComponent); // Override;
    var
      i       : Integer;
      aModule : TCustomModule;
    begin
      For i := 0 To ModuleCount - 1 Do
      Begin
        aModule := Module[ i];
        If aModule.Owner = Root
        Then aProc( aModule);
      End;
    End;

    Procedure   TDrawSheet.DefineProperties( aFiler: TFiler); // Override;
    Begin
      Inherited DefineProperties( aFiler);
      aFiler.DefineProperty( 'Connections', ReadConnections, WriteConnections, True);
    End;

    Procedure   TDrawSheet.WriteConnections( aWriter: TWriter);
    Var
      i : Integer;
    Begin
      With aWriter Do
      Begin
        WriteListBegin;
        For i := 0 To ConnectionCount - 1 Do
          WriteString(
            // New code, name based.
            Format(
              '%s.%s %s.%s',
              [
                Source     [ i].Owner.Name,
                Source     [ i]      .Name,
                Destination[ i].Owner.Name,
                Destination[ i]      .Name
              ]
            )

{
            Old code, index based

              ********************************************************
              ********************************************************
              ***                                                  ***
              ***   do not delete this, the program still hass     ***
              ***   support to read old format patches, so this    ***
              ***   is documentation ...                           ***
              ***                                                  ***
              ********************************************************
              ********************************************************

            Format(
              '%d.%d %d.%d',
              [
                Source     [ i].Owner.ComponentIndex,
                Source     [ i]      .ComponentIndex,
                Destination[ i].Owner.ComponentIndex,
                Destination[ i]      .ComponentIndex
              ]
            )
}

          );
        WriteListEnd;
      End;
    End;

    Procedure   TDrawSheet.ReadConnections( aReader: TReader);
    Begin
      ClearConnectionFixups;
      With aReader Do
      Begin
        ReadListBegin;
        While Not EndOfList Do
          AddConnectionFixup( ReadString);
        ReadListEnd;
      End;
    End;

    Procedure   TDrawSheet.Loaded; // Override;
    Begin
      FixupConnections;
      Inherited Loaded;
    End;

    //
    //
    // Reader / Writer / Persistence - stuff
    //
    ///////////////////////////////////////////////////////////////////////////


{  Public }

    Constructor TDrawSheet.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FPen               := TPen.Create;
      ControlStyle       := ControlStyle + [ csReplicatable, csOpaque];
      FCanvas            := TCanvas.Create;
      Align              := alClient;
      FModules           := TList.Create;
      FConnections       := TConnections.Create( Self);
      OnMouseDown        := DoMouseDown;
      OnMouseUp          := DoMouseUp;
      OnMouseMove        := DoMouseMove;
      FDoDrawConnections := True;
      FSlotActive        := True;
    End;

    Destructor  TDrawSheet.Destroy; // Override;
    Begin
      FConnections.Free;
      FModules    .Free;
      FCanvas     .Free;
      FPen        .Free;
      If Assigned( FConnectionFixups)
      Then FConnectionFixups.Free;
      Inherited Destroy;
    End;

    Procedure   TDrawSheet.Execute; // Virtual;
    Var
      i : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
        Module[ i].Execute;
      FConnections.Execute;
    End;

    Procedure   TDrawSheet.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Virtual;
    Var
      i : Integer;
    Begin
      WriteLn( aFile, Indent( anInd, ClassName));
      If aDepth > 0
      Then Begin
        WriteLn( aFile, Indent( anInd, 'Begin'));
        For i := 0 To ModuleCount - 1 Do
          Module[ i].Dump( aFile, anInd + 1, aDepth - 1);
        WriteLn( aFile, Indent( anInd, 'End'));
      End;
    End;

    Procedure   TDrawSheet.AcceptMidiShort( aShortMidi: TShortMidi);
    Var
      i : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
        If Module[ i] Is TMidiModule
        Then TMidiModule( Module[ i]).AcceptMidiShort( aShortMidi);
    End;

    Procedure   TDrawSheet.AcceptMidiLong( Const aLongMidi: TLongMidi);
    Var
      i : Integer;
    Begin
        For i := 0 To ModuleCount - 1 Do
          If Module[ i] Is TMidiModule
          Then TMidiModule( Module[ i]).AcceptMidiLong( aLongMidi);
    End;

    Function    TDrawSheet.CreateModule( aModuleClass: TModuleClass): TCustomModule;
    Begin
      GAutoCreation := True;
      Try
        Result := aModuleClass.Create( Self);
        With Result Do
        Begin
          GDragged    := Result;
          GAnchor     := CenterPoint;
          GDistance   := Point( 0, 0);
          BringToFront;
          Selected := True;
          UnselectPeers;
          HandleUpdateMove( GDistance);
        End;
      Finally
        GAutoCreation := False;
      End;
    End;

    Procedure   TDrawSheet.FreeModule( anIndex: Integer);
    Begin
      Module[ anIndex].Free;
    End;

    Procedure   TDrawSheet.FreeModules;
    Var
      i : Integer;
    Begin
      With FModules Do
        For i := Count - 1 Downto 0 Do
          FreeModule( i);
    End;

    Procedure   TDrawSheet.FreeSelectedModules;
    Var
      i : Integer;
    Begin
      With FModules Do
        For i := Count - 1 Downto 0 Do
          If Module[ i].Selected
          Then FreeModule( i);
    End;


    Procedure   TDrawSheet.CopyModulesToClipBoard;
    Var
      i          : Integer;
      aDrawSheet : TDrawSheet;
      aModule    : TCustomModule;
      aSrcTag    : LongInt;
      aDstTag    : LongInt;
    Begin
      aDrawSheet := TDrawSheet.Create( Nil);
      Try
        With FModules Do
          For i := 0 To Count - 1 Do
            If Module[ i].Selected
            Then Begin
              aModule := ( TModuleClass( Module[ i].ClassType)).Create( aDrawSheet);
              Module[ i].Tag := LongInt( aModule);
              With aModule Do
              Begin
                Assign( Module[ i]);
                Module[ i].MarkConnections;
              End;
            End
            Else Module[ i].Tag := 0;
        For i := 0 To ConnectionCount - 1 Do
          With Connection[ i] Do
            If FMarked
            Then Begin
              aSrcTag := ( Source.     Owner As TCustomModule).Tag;
              aDstTag := ( Destination.Owner As TCustomModule).Tag;
              If ( aSrcTag <> 0) And ( aDstTag <> 0)
              Then Begin
                With aDrawSheet Do
                Begin
                  FConnections.Connect(
                    TCustomModule( aSrcTag).Components[ Self.Source     [ i].ComponentIndex] As TConnector,
                    TCustomModule( aDstTag).Components[ Self.Destination[ i].ComponentIndex] As TConnector
                  );
                End;
              End;
            End;
        ClipBoard.SetComponent( aDrawSheet);
      Finally
        aDrawSheet.Free;
      End;
    End;

    Procedure   TDrawSheet.PasteModulesFromClipBoard;
    Var
      aDummy    : TComponent;
      aComponent: TComponent;
      aModule   : TCustomModule;
      i         : Integer;
      aSrcTag   : Integer;
      aDstTag   : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
        Module[ i].Selected := False;
      aDummy     := TComponent.Create( Application);
      aComponent := ClipBoard.GetComponent( aDummy, aDummy);
      Try
        If aComponent Is TDrawSheet
        Then Begin
          With TDrawSheet( aComponent) Do
          Begin
            GAutoCreation := True;
            For i := 0 To ModuleCount - 1 Do
            Begin
              aModule := ( TModuleClass( Module[ i].ClassType)).Create( Nil);
              With aModule Do
              Begin
                Assign( Module[ i]);
                Selected    := True;
                FInCreation := False;
              End;
              Module[ i].Tag := Integer( aModule);
              Self.InsertComponent( aModule);
            End;
            GAutoCreation := False;
            // Still working with TDrawSheet( aComponent), not with Self ...
            For i := 0 To ConnectionCount - 1 Do
              With Connection[ i] Do
              Begin
                aSrcTag := ( Source.     Owner As TCustomModule).Tag;
                aDstTag := ( Destination.Owner As TCustomModule).Tag;
                If ( aSrcTag <> 0) And ( aDstTag <> 0)
                Then
                  Self.FConnections.Connect(
                    TCustomModule( aSrcTag).Components[ Source     .ComponentIndex] As TConnector,
                    TCustomModule( aDstTag).Components[ Destination.ComponentIndex] As TConnector
                  );
              End;
          End;
        End
        Else ShowMessage( 'Clipboard has no modules.');
      Finally
        aDummy.Free;
        Invalidate;
      End;
    End;

    Procedure   TDrawSheet.SelectAllModules;
    Var
      i : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
       Module[ i].Selected := True;
    End;

    Procedure   TDrawSheet.InvertModuleSelection;
    Var
      i : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
       With Module[ i] Do
         Selected := Not Selected;
    End;

    Function    TDrawSheet.IndexOfModule( aModule: TCustomModule): Integer;
    Begin
      Result := FModules.IndexOf( aModule);
    End;

    Function    TDrawSheet.ModuleCount: Integer;
    Begin
      If Assigned( FModules)
      Then Result := FModules.Count
      Else Result := 0;
    End;

    Function    TDrawSheet.ConnectionCount: Integer;
    Begin
      If Assigned( FConnections)
      Then Result := FConnections.Count
      Else Result := 0;
    End;

    Procedure   TDrawSheet.VisualMaintainance; // Virtual;
    Var
      i : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
        Module[ i].VisualMaintainance;
    End;


{ ========
  TSlot = Class( TScrollBox)
  Private
      FOnRightClick       : TNotifyEvent;
      FOnCountChange      : TNotifyCountChange;
      FPrivateCountChange : TNotifyCountChange;
      FOnConnectionMade   : TNotifyConnectionChange;
      FOnConnectionLost   : TNotifyConnectionChange;
      FOnMidiOutShort     : TMidiOutShortProc;
      FOnMidiOutLong      : TMidiOutLongProc;
      FDrawSheet          : TDrawSheet;
      FFileName           : String;
      FActive             : Boolean;
  Protected
    Property    OnPrivateCountChange : TNotifyCountChange Read FPrivateCountChange Write FPrivateCountChange;
  Published
    Property    OnRightClick     : TNotifyEvent            Read FOnRightClick     Write FOnRightClick;
    Property    OnCountChange    : TNotifyCountChange      Read FOnCountChange    Write FOnCountChange;
    Property    OnConnectionMade : TNotifyConnectionChange Read FOnConnectionMade Write FOnConnectionMade;
    Property    OnConnectionLost : TNotifyConnectionChange Read FOnConnectionLost Write FOnConnectionLost;
    Property    OnMidiOutShort   : TMidiOutShortProc       Read FOnMidiOutShort   Write FOnMidiOutShort;
    Property    OnMidiOutLong    : TMidiOutLongProc        Read FOnMidiOutLong    Write FOnMidiOutLong;
    Property    SlotFileName     : String                  Read FFileName;
    Property    Active           : Boolean                 Read FActive           Write SetActive;
  Private
}

    Procedure   TSlot.PerformCountChange;
    Begin
      If Not ( csDestroying In ComponentState)
      Then Begin
        If Assigned( FPrivateCountChange)
        Then FPrivateCountChange( Self, ModuleCount);
        If Assigned( FOnCountChange)
        Then FOnCountChange( Self, ModuleCount);
      End;
    End;

    Procedure   TSlot.SendDblClick( Sender: TObject);
    Begin
      If Assigned( OnDblClick)
      Then OnDblClick( TObject( Sender));
    End;

    Procedure   TSlot.SendRightClick( Sender: TObject);
    Begin
      If Assigned( OnRightClick)
      Then OnRightClick( TObject( Sender));
    End;

    Procedure   TSlot.UMPaint( Var aMessage: TUMPaint); // Message UM_PAINT;
    Begin
      If HandleAllocated
      Then Begin
        InvalidateRect( Handle, aMessage.Data, True);
        FreeMem( aMessage.Data, SizeOf( TRect));
      End;
    End;

    Procedure   TSlot.PerformMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Begin
      If Assigned( FOnMidiOutShort) And Active
      Then FOnMidiOutShort( aSender, aShortMidi);
    End;

    Procedure   TSlot.PerformMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
    Begin
      If Assigned( FOnMidiOutLong) And Active
      Then FOnMidiOutLong( aSender, aLongMidi);
    End;

    Procedure   TSlot.SetActive( aValue: Boolean);
    Begin
      If FActive <> aValue
      Then Begin
        FActive := aValue;
      End;
    End;

{  Protected }

    Procedure   TSlot.ConnectionMade( aSource, aDestination: TConnector); // Virtual;
    Begin
      If Assigned( FOnConnectionMade)
      Then FOnConnectionMade( Self, aSource, aDestination);
    End;

    Procedure   TSlot.ConnectionLost( aSource, aDestination: TConnector);  // Virtual;
    Begin
      If Assigned( FOnConnectionLost)
      Then FOnConnectionLost( Self, aSource, aDestination);
    End;

{  Public }

    Constructor TSlot.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      BorderStyle := bsNone;
      FFileName   := '';
      FDrawSheet  := TDrawSheet.Create( Self);
      With FDrawSheet Do
      Begin
        Parent := Self;
        FSlot  := Self;
      End;
      Active := True;
    End;

    Procedure   TSlot.Execute; // Virtual;
    Begin
      FDrawSheet.Execute;
    End;

    Destructor  TSlot.Destroy; // Override;
    Begin
      Inherited Destroy;
    End;

    Procedure   TSlot.Save; // Virtual;
    Begin
      SaveAs( SlotFileName);
    End;

    Procedure   TSlot.SaveAs( Const aFileName: String); // Virtual;
    Var
      Sb : TStream;
      St : TStream;
    Begin
      FDrawSheet.SyncModules;
      St := Nil;
      Sb := TMemoryStream.Create;
      Try
        With Sb Do
        Begin
          WriteComponentRes( '', FDrawSheet);
          Seek( 0, 0);
        End;
        St := TFileStream.Create( aFileName, fmCreate);
        ObjectResourceToText( Sb, St);
      Finally
        If Assigned( St)
        Then St.Free;
        Sb.Free;
      End;
      FFileName := aFileName;
    End;

    Procedure   TSlot.Load( Const aFileName: String); // Virtual;
    Var
      aDrawSheet: TDrawSheet;
      Sb        : TStream;
      St        : TStream;
    Begin
      Sb := TMemoryStream.Create;
      St := TFileStream.Create( aFileName, fmOpenRead);
      Try
        ObjectTextToResource( St, Sb);
        aDrawSheet := TDrawSheet.Create( Self);
        With Sb Do
        Begin
          Seek( 0, 0);
          ReadComponentRes( aDrawSheet);
        End;
        FDrawSheet.Free;
        FDrawSheet := aDrawSheet;
        With FDrawSheet Do
        Begin
          Parent := Self;
          FSlot  := Self;
          If ModuleCount > 0
          Then ScrollModuleInView( Module[ 0]);
        End;
      Finally
        Sb.Free;
        St.Free;
      End;
      FFileName := aFileName;
    End;

    Procedure   TSlot.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Virtual;
    Begin
      WriteLn( aFile, Indent( anInd, 'Slot'));
      If Assigned( FDrawSheet)
      Then FDrawSheet.Dump( aFile, anInd, aDepth)
    End;

    Procedure   TSlot.AcceptMidiShort( aShortMidi: TShortMidi);
    Begin
      If Assigned( FDrawSheet) And Active
      Then FDrawSheet.AcceptMidiShort( aShortMidi);
    End;

    Procedure   TSlot.AcceptMidiLong( Const aLongMidi: TLongMidi);
    Begin
      If Assigned( FDrawSheet) And Active
      Then FDrawSheet.AcceptMidiLong( aLongMidi);
    End;

    Function    TSlot.CreateModule( aModuleClass: TModuleClass): TCustomModule;
    Begin
      If Assigned( FDrawSheet)
      Then Result := FDrawSheet.CreateModule( aModuleClass)
      Else Result := Nil;
    End;

    Procedure   TSlot.FreeModule( anIndex: Integer);
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.FreeModule( anIndex);
    End;

    Procedure   TSlot.FreeModules;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.FreeModules;
    End;

    Procedure   TSlot.FreeSelectedModules;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.FreeSelectedModules;
    End;

    Procedure   TSlot.CopyModulesToClipBoard;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.CopyModulesToClipBoard;
    End;

    Procedure   TSlot.PasteModulesFromClipBoard;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.PasteModulesFromClipBoard;
    End;

    Procedure   TSlot.SelectAllModules;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.SelectAllModules;
    End;

    Procedure   TSlot.InvertModuleSelection;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.InvertModuleSelection;
    End;

    Function    TSlot.ModuleCount: Integer;
    Begin
      If Assigned( FDrawSheet)
      Then Result := FDrawSheet.ModuleCount
      Else Result := 0;
    End;

    Procedure   TSlot.ScrollModuleInView( aModule: TCustomModule);
    var
      aRect: TRect;
    begin
      if Assigned( aModule)
      Then Begin
        aRect := aModule.ClientRect;
        Dec( aRect.Left  , HorzScrollBar.Margin);
        Inc( aRect.Right , HorzScrollBar.Margin);
        Dec( aRect.Top   , VertScrollBar.Margin);
        Inc( aRect.Bottom, VertScrollBar.Margin);
        aRect.TopLeft     := ScreenToClient( aModule.ClientToScreen( aRect.TopLeft    ));
        aRect.BottomRight := ScreenToClient( aModule.ClientToScreen( aRect.BottomRight));
        if aRect.Left < 0
        then
          with HorzScrollBar do
            Position := Position + aRect.Left
        else if aRect.Right > ClientWidth
        then begin
          if aRect.Right - aRect.Left > ClientWidth
          then aRect.Right := aRect.Left + ClientWidth;
          with HorzScrollBar do
            Position := Position + aRect.Right - ClientWidth;
        end;
        if aRect.Top < 0
        then with VertScrollBar do
          Position := Position + aRect.Top
        else if aRect.Bottom > ClientHeight
        then begin
          if aRect.Bottom - aRect.Top > ClientHeight
          then aRect.Bottom := aRect.Top + ClientHeight;
          with VertScrollBar do
            Position := Position + aRect.Bottom - ClientHeight;
        end;
      End;
    End;

    Procedure   TSlot.VisualMaintainance; // Virtual;
    Begin
      FDrawSheet.VisualMaintainance;
    End;



{ ========
  TModulePageControl = Class( TPageControl)
  Private
      FOnSlotCountChange   : TNotifyCountChange;
      FOnModuleCountChange : TNotifyCountChange;
      FOnConnectionMade    : TNotifyConnectionChange;
      FOnConnectionLost    : TNotifyConnectionChange;
      FOnActiveSlotChange  : TNotifyActiveSlotChange;
      FOnMidiOutShort      : TMidiOutShortProc;
      FOnMidiOutLong       : TMidiOutLongProc;
      FOnDetailClick       : TNotifyEvent;
      FOnDetailRightClick  : TNotifyEvent;
      FOnDetailDblClick    : TNotifyEvent;
      FPageCount           : Integer;
      FLockCount           : Integer;
  Public
    Property    ActiveSlot: TSlot Read GetActiveSlot;
  Published
    Property    CurvedLines         : Boolean                 Read GetCurvedLines       Write SetCurvedLines;
    Property    OnSlotCountChange   : TNotifyCountChange      Read FOnSlotCountChange   Write FOnSlotCountChange;
    Property    OnModuleCountChange : TNotifyCountChange      Read FOnModuleCountChange Write FOnModuleCountChange;
    Property    OnConnectionMade    : TNotifyConnectionChange Read FOnConnectionMade    Write FOnConnectionMade;
    Property    OnConnectionLost    : TNotifyConnectionChange Read FOnConnectionLost    Write FOnConnectionLost;
    Property    OnActiveSlotChange  : TNotifyActiveSlotChange Read FOnActiveSlotChange  Write FOnActiveSlotChange;
    Property    OnMidiOutShort      : TMidiOutShortProc       Read FOnMidiOutShort      Write FOnMidiOutShort;
    Property    OnMidiOutLong       : TMidiOutLongProc        Read FOnMidiOutLong       Write FOnMidiOutLong;
    Property    OnDetailClick       : TNotifyEvent            Read FOnDetailClick       Write FOnDetailClick;
    Property    OnDetailRightClick  : TNotifyEvent            Read FOnDetailRightClick  Write FOnDetailRightClick;
    Property    OnDetailDblClick    : TNotifyEvent            Read FOnDetailDblClick    Write FOnDetailDblClick;
  Private
}

    Function    TModulePageControl.GetActiveSlot: TSlot;
    Var
      i : Integer;
    Begin
      Result := Nil;
      If Assigned( ActivePage)
      Then
        With ActivePage Do
        Begin
          For i := 0 To ControlCount - 1 Do
            If Controls[ i] Is TSlot
            Then Begin
              Result := TSlot( Controls[ i]);
              Break;
            End;
        End;
    End;

    Procedure   TModulePageControl.DoPrivateCountChange( Sender: TObject; aCount: Integer);
    // Handles Count changes for Modules
    Begin
      If Sender Is TSlot
      Then PerformModuleCountChange( TSlot( Sender));
    End;

    Procedure   TModulePageControl.DoPrivateConnectionMade( aSender: TObject; aSource, aDestination: TConnector);
    Begin
      PerformConnectionMade( aSender, aSource, aDestination);
    End;

    Procedure   TModulePageControl.DoPrivateConnectionLost( aSender: TObject; aSource, aDestination: TConnector);
    Begin
      PerformConnectionLost( aSender, aSource, aDestination);
    End;

    Procedure   TModulePageControl.DoMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Begin
      If Assigned( FOnMidiOutShort) And Not( csDestroying In ComponentState)
      Then FOnMidiOutShort( aSender, aShortMidi);
    End;

    Procedure   TModulePageControl.DoMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
    Begin
      If Assigned( FOnMidiOutLong) And Not( csDestroying In ComponentState)
      Then FOnMidiOutLong( aSender, aLongMidi);
    End;

    Procedure   TModulePageControl.PerformSlotCountChange;
    // Handles Count changes for Slots
    Begin
      If Assigned( FOnSlotCountChange)
      Then Begin
        If csDestroying In ComponentState
        Then FOnSlotCountChange( Self, 0)
        Else FOnSlotCountChange( Self, PageCount);
      End;
      PerformModuleCountChange( ActiveSlot);
    End;

    Procedure   TModulePageControl.PerformModuleCountChange( Sender: TSlot);
    // Handles Count changes for Modules
    Begin
      If Assigned( FOnModuleCountChange)
      Then Begin
        If csDestroying In ComponentState
        Then FOnModuleCountChange( Self, 0)
        Else Begin
          If Assigned( Sender)
          Then FOnModuleCountChange( Sender, Sender.ModuleCount)
          Else FOnModuleCountChange( Self  , 0                 );
        End;
      End;
    End;

    Procedure   TModulePageControl.PerformConnectionMade( aSender: TObject; aSource, aDestination: TConnector);
    Begin
      If Assigned( FOnConnectionMade) And Not ( csDestroying In ComponentState)
      Then FOnConnectionMade( aSender, aSource, aDestination);
    End;

    Procedure   TModulePageControl.PerformConnectionLost( aSender: TObject; aSource, aDestination: TConnector);
    Begin
      If Assigned( FOnConnectionLost) And Not ( csDestroying In ComponentState)
      Then FOnConnectionLost( aSender, aSource, aDestination);
    End;

    Procedure   TModulePageControl.PerformSlotChange( aSender: TObject);
    Begin
      If Assigned( FOnActiveSlotChange)
      Then
        If ActiveSlot = Nil
        Then FOnActiveSlotChange( Nil, '')
        Else FOnActiveSlotChange( ActiveSlot, ActivePage.Caption);
    End;

    Procedure   TModulePageControl.PerformClick( aSender: TObject);
    Begin
      If Assigned( OnDetailClick)
      Then OnDetailClick( aSender);
    End;

    Procedure   TModulePageControl.PerformRightClick( aSender: TObject);
    Begin
      If Assigned( OnDetailRightClick)
      Then OnDetailRightClick( aSender);
    End;

    Procedure   TModulePageControl.PerformDblClick( aSender: TObject);
    Begin
      If Assigned( OnDetailDblClick)
      Then OnDetailDblClick( aSender);
    End;

    Procedure   TModulePageControl.Lock;
    Begin
      Inc( FLockCount);
    End;

    Procedure   TModulePageControl.UnLock;
    Begin
      If FLockCount > 0
      Then Dec( FLockCount);
    End;

    Function    TModulePageControl.GetCurvedLines: Boolean;
    Begin
      Result := GUseCurvedLines;
    End;

    Procedure   TModulePageControl.SetCurvedLines( aValue: Boolean);
    Begin
      If GUseCurvedLines <> aValue
      Then Begin
        GUseCurvedLines := aValue;
        If Assigned( ActiveSlot)
        Then ActiveSlot.Invalidate;
      End;
    End;

{   Protected }

    Procedure   TModulePageControl.Change; // Override;
    Begin
      Inherited Change;
      PerformModuleCountChange( ActiveSlot);
    End;

{  Public }

    Constructor TModulePageControl.Create( anOwner: TComponent);  // Override;
    Begin
      Inherited Create( anOwner);
      OnChange := PerformSlotChange;
      HotTrack := True;
    End;

    Procedure   TModulePageControl.Execute; // Virtual;
    Var
      i : Integer;
      j : Integer;
    Begin
      If ( FLockCount = 0) And Not( csDestroying In ComponentState)
      Then
        For i := 0 To PageCount - 1 Do
          With Pages[ i] Do
            For j := 0 To ControlCount - 1 Do
              If Controls[ j] Is TSlot
              Then TSlot( Controls[ j]).Execute;
    End;

    Procedure   TModulePageControl.Save; // Virtual;
    Begin
      If Assigned( ActiveSlot)
      Then ActiveSlot.Save;
    End;

    Procedure   TModulePageControl.SaveAll; // Virtual;
    Var
      i : Integer;
      j : Integer;
    Begin
      For i := 0 To PageCount - 1 Do
        With Pages[ i] Do
          For j := 0 To ControlCount - 1 Do
            If Controls[ j] Is TSlot
            Then TSlot( Controls[ j]).Save;
    End;

    Procedure   TModulePageControl.SaveAs( Const aFileName: String); // Virtual;
    Begin
      If Assigned( ActiveSlot)
      Then Begin
        ActiveSlot.SaveAs( aFileName);
        ActivePage.Caption := CleanFileName( aFileName);
        PerformSlotChange( Self);
      End;
    End;

    Procedure   TModulePageControl.Load( Const aFileName: String); // Virtual;
    Begin
      Lock;
      Try
        AddSlot( CleanFileName( aFileName));
        If Assigned( ActiveSlot)
        Then Begin
          Try
            ActiveSlot.Load( aFileName);
            PerformModuleCountChange( ActiveSlot);
          Except
            On Exception Do
            Begin
              ActivePage.Free;
              Raise;
            End;
          End;
        End;
      Finally
        UnLock;
      End;
    End;

    Procedure   TModulePageControl.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Virtual;
    Var
      i : Integer;
      j : Integer;
    Begin
      WriteLn( aFile, Indent( anInd, ClassName));
      WriteLn( aFile, Indent( anInd, 'Begin'));
      If aDepth > 0
      Then Begin
        For i := 0 To PageCount - 1 Do
          With Pages[ i] Do
            For j := 0 To ControlCount - 1 Do
              If Controls[ j] Is TSlot
              Then TSlot( Controls[ j]).Dump( aFile, anInd + 1, aDepth - 1);
        WriteLn( aFile, Indent( anInd, 'End'));
      End;
    End;

    Procedure   TModulePageControl.AddSlot( Const aName: String);
    Var
      T : TTabSheet;
      S : String;
    Begin
      If aName = ''
      Then Begin
        Inc( FPageCount);
        S := Format( 'Slot%d', [ FPageCount]);
      End
      Else S := aName;
      T := TTabSheet.Create( Self);
      With T Do
      Begin
        Caption     := S;
        PageControl := Self;
        With TSlot.Create( T) Do
        Begin
          Parent               := T;
          FFileName            := ApplicationPath + S + DefaultSlotExt;
          Align                := alClient;
          Color                := clBlack;
          BorderStyle          := bsNone;
          OnPrivateCountChange := DoPrivateCountChange;
          OnConnectionMade     := DoPrivateConnectionMade;
          OnConnectionLost     := DoPrivateConnectionLost;
          OnMidiOutShort       := DoMidiOutShort;
          OnMidiOutLong        := DoMidiOutLong;
          OnClick              := PerformClick;
          OnRightClick         := PerformRightClick;
          OnDblClick           := PerformDblClick;
        End;
        Visible := True;
      End;
      ActivePage := T;
      PerformSlotCountChange;
      PerformSlotChange( Self);
    End;

    Procedure   TModulePageControl.AcceptMidiShort( aShortMidi: TShortMidi);
    Var
      i, j : Integer;
    Begin
      For i := 0 To PageCount - 1 Do
        With Pages[ i] Do
          For j := 0 To ControlCount - 1 Do
            If Controls[ j] Is TSlot
            Then TSlot( Controls[ j]).AcceptMidiShort( aShortMidi);
    End;

    Procedure   TModulePageControl.AcceptMidiLong( Const aLongMidi: TLongMidi);
    Var
      i, j : Integer;
    Begin
      For i := 0 To PageCount - 1 Do
        With Pages[ i] Do
          For j := 0 To ControlCount - 1 Do
            If Controls[ j] Is TSlot
            Then TSlot( Controls[ j]).AcceptMidiLong( aLongMidi);
    End;

    Procedure   TModulePageControl.FreeCurrentSlot;
    Begin
      Lock;
      Try
        If Assigned( ActivePage)
        Then Begin
          ActivePage.Free;
          If PageCount > 0
          Then ActivePage := Pages[ 0];
        End;
      Finally
        UnLock;
      End;
      PerformSlotCountChange;
      PerformSlotChange( Self);
    End;

    Procedure   TModulePageControl.FreeAllSlots;
    Begin
      While Assigned( ActivePage) Do
        FreeCurrentSlot;
    End;

    Function    TModulePageControl.CreateModule( aModuleClass: TModuleClass): TCustomModule;
    Begin
      If Assigned( ActiveSlot)
      Then Begin
        Result := ActiveSlot.CreateModule( aModuleClass);
        With Result Do
        Begin
          Top  := 0;
          Left := 0;
        End;
      End
      Else Result := Nil;
    End;

    Procedure   TModulePageControl.FreeModule( anIndex: Integer);
    Begin
      Lock;
      Try
        If Assigned( ActiveSlot)
        Then ActiveSlot.FreeModule( anIndex);
      Finally
        UnLock;
      End;
    End;

    Procedure   TModulePageControl.FreeModules;
    Begin
      Lock;
      Try
        If Assigned( ActiveSlot)
        Then ActiveSlot.FreeModules;
      Finally
        UnLock;
      End;
    End;

    Procedure   TModulePageControl.FreeSelectedModules;
    Begin
      Lock;
      Try
        If Assigned( ActiveSlot)
        Then ActiveSlot.FreeSelectedModules;
      Finally
        UnLock;
      End;
    End;

    Procedure   TModulePageControl.CopyModulesToClipBoard;
    Begin
      If Assigned( ActiveSlot)
      Then ActiveSlot.CopyModulesToClipBoard;
    End;

    Procedure   TModulePageControl.PasteModulesFromClipBoard;
    Begin
      Lock;
      Try
        If Assigned( ActiveSlot)
        Then ActiveSlot.PasteModulesFromClipBoard;
      Finally
        UnLock;
      End;
    End;

    Procedure   TModulePageControl.SelectAllModules;
    Begin
      If Assigned( ActiveSlot)
      Then ActiveSlot.SelectAllModules;
    End;

    Procedure   TModulePageControl.InvertModuleSelection;
    Begin
      If Assigned( ActiveSlot)
      Then ActiveSlot.InvertModuleSelection;
    End;

    Procedure   TModulePageControl.ShakeCables;
    Begin
      GBezierAngle1    := 15 - Random( 31);
      GBezierStrength1 := 30 + Random( 51);
      GBezierAngle2    := 15 - Random( 31);
      GBezierStrength2 := 30 + Random( 51);
      If Assigned( ActiveSlot)
      Then ActiveSlot.Invalidate;
    End;

    Function    TModulePageControl.ModuleCount: Integer;
    Begin
      If Assigned( ActiveSlot)
      Then Result := ActiveSlot.ModuleCount
      Else Result := 0;
    End;

    Procedure   TModulePageControl.VisualMaintainance; // Virtual;
    Begin
      If Assigned( ActiveSlot)
      Then ActiveSlot.VisualMaintainance;
    End;


{ ========
  TModuleButton = Class( TSpeedButton)
  Private
      FPageName    : String;
      FModuleClass : TModuleClass;
  Public
    Property    PageName   : String       Read FPageName;
    Property    ModuleClass: TModuleClass Read FModuleClass;
}

    Procedure   TModuleButton.LoadGlyph;
    Begin
      With ModuleClass Do
        Glyph := CreateBitmapFromFile( Copy( ClassName, 2, Length( ClassName)) + '.bmp');
    End;

    Procedure   TModuleButton.Initialize( Const aPageName: String; aModuleClass: TModuleClass);
    Begin
      FPageName    := aPageName;
      FModuleClass := aModuleClass;
      LoadGlyph;
    End;


{ ========
  TModuleButtonClick = Procedure( aSender: TObject; aModuleClass: TModuleClass) Of Object;

  TmoduleSelector = Class( TPageControl)
  Private
      FRegisteredModules   : TList;
      FOnModuleButtonClick : TModuleButtonClick;
      FButtonsEnabled      : Boolean;
  Published
    Property    OnModuleButtonClick : TModuleButtonClick Read FonModuleButtonClick Write FOnModuleButtonClick;
    Property    ButtonsEnabled      : Boolean            Read FButtonsEnabled      Write SetButtonsEnabled;
  Private
}

    Procedure   TmoduleSelector.SetButtonsEnabled( aValue: Boolean);
    Var
      i : Integer;
    Begin
      If aValue <> FButtonsEnabled
      Then Begin
        FButtonsEnabled := aValue;
        With FRegisteredModules Do
          For i := 0 To Count - 1 Do
            TModuleButton( Items[ i]).Enabled := FButtonsEnabled;
      End;
    End;

    Function    TmoduleSelector.AddPage( Const aName: String): TTabSheet;
    Begin
      Result := TTabSheet.Create( Self);
      With Result Do
      Begin
        Visible     := True;
        Caption     := aName;
        PageControl := Self;
      End;
    End;

    Function    TmoduleSelector.PageButtonCount( Const aName: String): Integer;
    Var
      i : Integer;
    Begin
      Result := 0;
      If PageExists( aName)
      Then
        With PageByName( aName) Do
          For i := 0 To ControlCount - 1 Do
            If Controls[ i] Is TModuleButton
            Then Inc( Result);
    End;

    Function    TmoduleSelector.PageExists( Const aName: String): Boolean;
    Begin
      Result := PageByName( aName) <> Nil;
    End;

    Function    TmoduleSelector.PageByName( Const aName: String): TTabSheet;
    Var
      i : Integer;
    Begin
      Result := Nil;
      For i := 0 To PageCount - 1 Do
        If Pages[ i].Caption = aName
        Then Begin
          Result := Pages[ i];
          Break;
        End;
    End;

    Function    TmoduleSelector.FindModulePage( aModuleClass: TModuleClass): TTabSheet;
    Var
      i : Integer;
    Begin
      Result := Nil;
      With FRegisteredModules Do
        For i := 0 To Count - 1 Do
          With TModuleButton( Items[ i]) Do
            If ModuleClass = aModuleClass
            Then Begin
              Result := PageByName( PageName);
              Break;
            End;

    End;

    Function    TmoduleSelector.FindModuleButton( aModuleClass: TModuleClass): TModuleButton;
    Var
      aPage : TTabSheet;
      i     : Integer;
    Begin
      aPage := FindModulePage( aModuleClass);
      Result := Nil;
      If Assigned( aPage)
      Then
        With aPage Do
        Begin
          For i := 0 To ControlCount - 1 Do
            If Controls[ i] Is TModuleButton
            Then
              With TModuleButton( Controls[ i]) Do
                If ModuleClass = aModuleClass
                Then Begin
                  Result := TModuleButton( Controls[ i]);
                  Break;
                End;
        End;

    End;

    Procedure   TmoduleSelector.DoModuleMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
    Begin
      If Assigned( FOnModuleButtonClick) And ( aSender Is TModuleButton)
      Then
        FOnModuleButtonClick(
          TModuleButton( aSender),
          TModuleButton( aSender).ModuleClass
        );
    End;

    Procedure   TmoduleSelector.FormatPage( aPage: TTabSheet);
    Var
      LastLeft : Integer;
      i        : Integer;
    Begin
      Assert(
        Assigned( aPage),
        'TModuleSelecteor.FormatPage : can''t format an unassigned <NIL> page.'
      );
      If PageButtonCount( aPage.Caption) = 0
      Then Begin
        aPage.Free;
        If PageCount > 0
        Then ActivePage := Pages[ 0];
      End
      Else
        With aPage Do
        Begin
          LastLeft := 0;
          For i := 0 To ControlCount - 1 Do
            If Controls[ i] Is TModuleButton
            Then
              With TModuleButton( Controls[ i]) Do
              Begin
                Left := LastLeft;
                Inc( LastLeft, Width);
              End;
        End;
    End;

{  Protected }

    Procedure   TmoduleSelector.Notification( aComponent: TComponent; anOperation: TOperation); // Override;
    Begin
      If aComponent Is TModuleButton
      Then Begin
        If anOperation = opInsert
        Then FRegisteredModules.Add   ( aComponent)
        Else FRegisteredModules.Remove( aComponent);
      End;
    End;

{   Public }

    Constructor TmoduleSelector.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FRegisteredModules := TList.Create;
      HotTrack := True;
    End;

    Destructor  TmoduleSelector.Destroy; // Override;
    Begin
      Inherited Destroy;
      FRegisteredModules.Free;
    End;

    Procedure   TmoduleSelector.RegisterModuleClass( Const aPageName: String; aModuleClass: TModuleClass);
    Var
      aPage         : TTabSheet;
      aModuleButton : TModuleButton;
    Begin

    // Uncomment (comment) the next section if you won't (want to) allow
    // for module duplicates.

    {
      If Assigned( FindModuleButton( aModuleClass))
        Then
          Raise
            EModule.
              CreateFmt(
                'A module type ''%s'' is already present on page ''%s''',
                [
                  aModuleClass.ClassName,
                  FindModulePage( aModuleClass).Caption
                ]
              );
    }

      aPage := PageByName( aPageName);
      If Not Assigned( aPage)
      Then aPage := AddPage( aPageName);
      ActivePage := aPage;
      ActivePage := Pages[ 0];
      aModuleButton := TModuleButton.Create( Self);
      With aModuleButton Do
      Begin
        Parent   := aPage;
        Left     := 0;
        Initialize( aPageName, aModuleClass);
        OnMouseDown := DoModuleMouseDown;
        FormatPage( aPage);
        Top      := 0;
        Flat     := True;
        Hint     := Copy( ModuleClass.ClassName, 2, Length( ClassName));
        ShowHint := True;
        Enabled  := FButtonsEnabled;
      End;
    End;

    Procedure   TmoduleSelector.UnregisterModuleClass( aModuleClass: TModuleClass);
    Var
      aModuleButton : TModuleButton;
      aPage         : TTabSheet;
    Begin
      aModuleButton := FindModuleButton( aModuleClass);
      aPage         := FindModulePage  ( aModuleClass);
      If Assigned( aModuleButton) And Assigned( aPage)
      Then Begin
        aModuleButton.Free;
        FormatPage( aPage); // This will remove the page if it's empty.
      End;
    End;



Initialization

  RegisterClasses(
    [
      TDrawSheet,
      TCustomModule,
      TModule,
      TModule2,
      TModule4,
      TModule8,
      TMidiModule
    ]
  );

  GRepeatTimer   := TTimer.Create( Application);
  BmKnob         := CreateBitMapFromFile( 'Knob.bmp'        );
  BmKnobMedium   := CreateBitMapFromFile( 'KnobMedium.bmp'  );
  BmKnobSmall    := CreateBitMapFromFile( 'KnobSmall.bmp'   );
  BmRoundGray    := CreateBitMapFromFile( 'RoundGray.bmp'   );
  BmRoundYellow  := CreateBitMapFromFile( 'RoundYellow.bmp' );
  BmRoundRed     := CreateBitMapFromFile( 'RoundRed.bmp'    );
  BmRoundBlue    := CreateBitMapFromFile( 'RoundBlue.bmp'   );
  BmSquareGray   := CreateBitMapFromFile( 'SquareGray.bmp'  );
  BmSquareYellow := CreateBitMapFromFile( 'SquareYellow.bmp');
  BmSquareRed    := CreateBitMapFromFile( 'SquareRed.bmp'   );
  BmSquareBlue   := CreateBitMapFromFile( 'SquareBlue.bmp'  );
  BmLedOn        := CreateBitMapFromFile( 'LedOn.bmp'       );
  BmLedOff       := CreateBitMapFromFile( 'LedOff.bmp'      );

Finalization

  BmKnob        .Free;
  BmKnobMedium  .Free;
  BmKnobSmall   .Free;
  BmRoundGray   .Free;
  BmRoundYellow .Free;
  BmRoundRed    .Free;
  BmRoundBlue   .Free;
  BmSquareGray  .Free;
  BmSquareYellow.Free;
  BmSquareRed   .Free;
  BmSquareBlue  .Free;
  BmLedOff      .Free;
  BmLedOn       .Free;

  UnregisterClasses(
    [
      TDrawSheet,
      TCustomModule,
      TModule,
      TModule2,
      TModule4,
      TModule8,
      TMidiModule
    ]
  );

end.


